from __future__ import annotations

import re
from typing import Any

import num2words

from .._morph import get_morph
from ..preprocess_utils import (
    NEGATIVE_NUMBER_PLACEHOLDER,
    PARAGRAPH_BREAK_PLACEHOLDER,
    normalize_ascii_quote_pairs,
)
from ..text_context import PUNCT_STRIP, normalize_context_token
from ._constants import ENTITY_DEFAULT_CASE, ENTITY_KEYWORDS, PREP_CASE, TIME_WORDS, VERB_CASE

SENTENCE_PUNCTUATION_PATTERN = re.compile(r"\s+([.,!?;:])")
POINT_NUMBER_SPACING_PATTERN = re.compile(r"(?<=\.) (?=\d)")
POINT_WORD_PATTERN = re.compile(r"(точка [а-яё]+)\. (?=[а-яё])", flags=re.IGNORECASE)
REPEATED_PUNCTUATION_PATTERN = re.compile(r"([,!?;:])\1+")
DOUBLE_DOT_PATTERN = re.compile(r"(?<!\.)\.\.(?!\.)")
COMMA_DOT_PATTERN = re.compile(r",\.")
DOT_COMMA_PATTERN = re.compile(r"\.,")
OPEN_BRACKET_SPACING_PATTERN = re.compile(r"([\(\[\{])\s+")
CLOSE_BRACKET_SPACING_PATTERN = re.compile(r"\s+([\)\]\}])")
WORD_RANGE_HYPHEN_PATTERN = re.compile(r"(?<=[A-Za-zА-Яа-яЁё]) - (?=[A-Za-zА-Яа-яЁё])")


def is_integer_token(token: str) -> bool:
    clean_token = token.strip(PUNCT_STRIP)
    if clean_token.startswith(NEGATIVE_NUMBER_PLACEHOLDER):
        clean_token = clean_token[len(NEGATIVE_NUMBER_PLACEHOLDER) :]
    return clean_token.isdigit()


def parse_integer_token(token: str) -> tuple[bool, str] | None:
    clean_token = token.strip(PUNCT_STRIP)
    is_negative = clean_token.startswith(NEGATIVE_NUMBER_PLACEHOLDER)
    if is_negative:
        clean_token = clean_token[len(NEGATIVE_NUMBER_PLACEHOLDER) :]
    if not clean_token.isdigit():
        return None
    return is_negative, clean_token


def build_number_token(
    token: str, clean_token: str, replacement: str, is_negative: bool
) -> str:
    source = (
        f"{NEGATIVE_NUMBER_PLACEHOLDER}{clean_token}" if is_negative else clean_token
    )
    result = token.replace(source, replacement, 1)
    if is_negative:
        return f"минус {result}"
    return result


def apply_genitive_plural_heuristic(word: str, pos: str | None) -> str:
    if pos not in {"ADJF", "PRTF"}:
        return word
    lower_word = word.lower()
    if lower_word.endswith("ые"):
        return word[:-2] + ("Ых" if word.isupper() else "ых")
    if lower_word.endswith("ие"):
        return word[:-2] + ("Их" if word.isupper() else "их")
    if lower_word.endswith("ый") or lower_word.endswith("ой"):
        return word[:-2] + ("Ого" if word.isupper() else "ого")
    if lower_word.endswith("ий") or lower_word.endswith("ее"):
        return word[:-2] + ("Его" if word.isupper() else "его")
    if lower_word.endswith("ая"):
        return word[:-2] + ("Ой" if word.isupper() else "ой")
    if lower_word.endswith("яя"):
        return word[:-2] + ("Ей" if word.isupper() else "ей")
    return word


def safe_inflect(
    parsed_word: Any,
    target_tags: set[str],
    fallback_word: str | None = None,
    pos_filter: set[str] | None = None,
) -> str:
    if fallback_word is None:
        fallback_word = parsed_word.word
    try:
        inflected = parsed_word.inflect(target_tags)
        if inflected:
            return inflected.word
    except Exception:
        pass
    pos = parsed_word.tag.POS
    if (
        (pos_filter is None or pos in pos_filter)
        and "gent" in target_tags
        and "plur" in target_tags
        and pos in {"ADJF", "PRTF"}
    ):
        heuristic_result = apply_genitive_plural_heuristic(parsed_word.word, pos)
        if heuristic_result != parsed_word.word:
            return heuristic_result
    return fallback_word


def inflect_unit_lemma(lemma: str, target_tags: set[str]) -> str:
    morph = get_morph()

    def pick_preferred_parse(word: str) -> Any | None:
        parsed = morph.parse(word)
        if not parsed:
            return None
        for candidate in parsed:
            if "NOUN" in candidate.tag:
                return candidate
        return parsed[0]

    if " " not in lemma:
        parsed = pick_preferred_parse(lemma)
        if not parsed:
            return lemma
        return safe_inflect(parsed, target_tags, fallback_word=lemma)

    parts = lemma.split()
    if len(parts) != 2:
        return lemma

    adj_text, noun_text = parts
    adj_parsed = pick_preferred_parse(adj_text)
    noun_parsed = pick_preferred_parse(noun_text)
    if not adj_parsed or not noun_parsed:
        return lemma

    noun_form = safe_inflect(noun_parsed, target_tags, fallback_word=noun_text)
    adj_tags = {
        tag
        for tag in target_tags
        if tag in {"nomn", "gent", "datv", "accs", "ablt", "loct", "sing", "plur"}
    }
    if adj_tags == {"gent", "sing"}:
        noun_gender = noun_parsed.tag.gender
        adj_tags = {"plur", "nomn"} if noun_gender == "femn" else {"gent", "plur"}
    elif "plur" not in adj_tags:
        noun_gender = noun_parsed.tag.gender
        if noun_gender:
            adj_tags.add(noun_gender)
    adj_form = safe_inflect(
        adj_parsed,
        adj_tags,
        fallback_word=adj_text,
        pos_filter={"ADJF", "PRTF"},
    )
    return f"{adj_form} {noun_form}"


def is_case_reliable_noun(parsed_word: Any) -> bool:
    return "NOUN" in parsed_word.tag and "Fixd" not in parsed_word.tag


def noun_number_form(n: int) -> str:
    if n == 0:
        return "many"
    last_two = n % 100
    last_digit = n % 10
    if last_two in {11, 12, 13, 14}:
        return "many"
    if last_digit == 1:
        return "one"
    if last_digit in {2, 3, 4}:
        return "few"
    return "many"


def should_consume_abbreviation_dot(tokens: list[str], dot_index: int) -> bool:
    if dot_index >= len(tokens) or tokens[dot_index] != ".":
        return False
    if dot_index + 1 >= len(tokens):
        return False
    next_token = tokens[dot_index + 1]
    if "\n" in next_token or PARAGRAPH_BREAK_PLACEHOLDER in next_token:
        return False
    if next_token in {",", ";", ":", ")", "]", "}", "»", '"', "”"}:
        return True
    stripped_next = next_token.strip(PUNCT_STRIP)
    if not stripped_next:
        return False
    first_char = stripped_next[:1]
    return first_char.islower() or first_char.isdigit()


def should_keep_decimal_unit_dot(rest: str) -> bool:
    stripped_rest = rest.lstrip()
    if not stripped_rest:
        return True
    if "\n" in rest[: len(rest) - len(stripped_rest)]:
        return True
    next_char = stripped_rest[:1]
    if next_char in ".!?…":
        return True
    return not (next_char.islower() or next_char.isdigit())


def inflect_numeral_string(num_str: str, case: str, gender: str | None = None) -> str:
    try:
        value = int(num_str)
    except ValueError:
        return num_str
    cases_map = {
        "nomn": "nominative",
        "gent": "genitive",
        "datv": "dative",
        "accs": "accusative",
        "ablt": "instrumental",
        "loct": "prepositional",
    }
    n2w_gender_map = {"masc": "masculine", "femn": "feminine", "neut": "neuter"}
    if case in cases_map:
        try:
            kwargs: dict[str, Any] = {"case": cases_map[case]}
            if gender == "plur":
                kwargs["plural"] = True
            elif gender in n2w_gender_map:
                kwargs["gender"] = n2w_gender_map[gender]
            return num2words.num2words(value, lang="ru", **kwargs)
        except Exception:
            pass
    try:
        words = num2words.num2words(value, lang="ru").split()
    except Exception:
        return num_str
    if case == "nomn" and gender is None:
        return " ".join(words)
    morph = get_morph()
    magnitudes = {
        "тысяча": "femn",
        "миллион": "masc",
        "миллиард": "masc",
        "триллион": "masc",
        "биллион": "masc",
    }

    def get_magnitude_gender(word_str: str) -> str | None:
        return magnitudes.get(morph.parse(word_str)[0].normal_form)

    inflected_words: list[str] = []
    for index, word in enumerate(words):
        parsed = morph.parse(word)
        p = parsed[0] if parsed else None
        if not p:
            inflected_words.append(word)
            continue
        current_gender = gender
        is_magnitude_self = get_magnitude_gender(word) is not None
        if not is_magnitude_self and index + 1 < len(words):
            mag_gender = get_magnitude_gender(words[index + 1])
            if mag_gender:
                current_gender = mag_gender
        if is_magnitude_self:
            current_gender = None
            if case == "nomn":
                inflected_words.append(word)
                continue
            if case == "accs" and word.lower() != "тысяча":
                inflected_words.append(word)
                continue
        tags = {case}
        if current_gender:
            tags.add(current_gender)
        inflected = p.inflect(tags)
        if inflected:
            inflected_words.append(inflected.word)
        else:
            if current_gender:
                inflected_case = p.inflect({case})
                if inflected_case:
                    inflected_words.append(inflected_case.word)
                    continue
            inflected_words.append(word)
    return " ".join(inflected_words)


def get_target_tags_for_number(
    num: int, case: str, noun_gender: str | None = None
) -> set[str]:
    del noun_gender
    form = noun_number_form(num)
    if case == "nomn":
        return (
            {"nomn", "sing"}
            if form == "one"
            else {"gent", "sing"} if form == "few" else {"gent", "plur"}
        )
    if case == "accs":
        return (
            {"accs", "sing"}
            if form == "one"
            else {"gent", "sing"} if form == "few" else {"gent", "plur"}
        )
    return {case, "sing"} if form == "one" else {case, "plur"}


def _get_preposition_before_number(tokens: list[str], idx: int) -> tuple[str, str] | None:
    max_prep_len = min(3, idx)
    for prep_len in range(max_prep_len, 1, -1):
        start = idx - prep_len
        prep_tokens = [normalize_context_token(token) for token in tokens[start:idx]]
        if not all(prep_tokens):
            continue
        phrase = " ".join(prep_tokens)
        if phrase in PREP_CASE:
            return phrase, PREP_CASE[phrase]
    for i in range(idx - 1, max(-1, idx - 3), -1):
        word_left = normalize_context_token(tokens[i])
        if word_left == "чем":
            break
        if word_left in PREP_CASE:
            return word_left, PREP_CASE[word_left]
    return None


def get_numeral_case(tokens: list[str], idx: int) -> str:
    morph = get_morph()
    is_range_start = idx < len(tokens) - 1 and tokens[idx + 1] in {"-", "–", "—"}

    def unit_hint(number_index: int) -> str | None:
        if number_index + 1 >= len(tokens):
            return None
        hint = normalize_context_token(tokens[number_index + 1])
        if hint == "°" and number_index + 2 < len(tokens):
            return hint + normalize_context_token(tokens[number_index + 2])
        return hint or None

    if idx > 1 and tokens[idx - 1] == "и":
        current_hint = unit_hint(idx)
        for back in range(idx - 2, max(-1, idx - 6), -1):
            if any(char in tokens[back] for char in ".!?;:"):
                break
            if is_integer_token(tokens[back]) and unit_hint(back) == current_hint:
                return get_numeral_case(tokens, back)
    if idx > 1 and tokens[idx - 1] in {"-", "–", "—"} and is_integer_token(tokens[idx - 2]):
        return get_numeral_case(tokens, idx - 2)

    if is_range_start and idx > 0:
        prev_token = tokens[idx - 1].strip(".,!?;:")
        if prev_token:
            p_prev = morph.parse(prev_token)[0]
            prev_case = "loct" if "loc2" in p_prev.tag else p_prev.tag.case
            if prev_case:
                if is_case_reliable_noun(p_prev):
                    if idx > 1:
                        prev_prev = normalize_context_token(tokens[idx - 2])
                        if prev_prev in {"в", "во", "о", "об", "при"}:
                            return "loct"
                        if prev_prev in PREP_CASE:
                            return PREP_CASE[prev_prev]
                    return prev_case
                if p_prev.tag.POS in {"ADJF", "PRTF"}:
                    return prev_case

    prep_match = _get_preposition_before_number(tokens, idx)
    if prep_match is not None:
        word_left, prep_case = prep_match
        if word_left in {"с", "со"}:
            for k in range(idx + 1, min(len(tokens), idx + 5)):
                if tokens[k].lower() in {"до", "по"}:
                    return "gent"
        if word_left in {"в", "на"}:
            for j in range(idx + 1, min(len(tokens), idx + 6)):
                if any(char in tokens[j] for char in ".!?;:,"):
                    break
                p = morph.parse(tokens[j])[0]
                if is_case_reliable_noun(p):
                    if p.tag.case == "loct" or "loc2" in p.tag:
                        return "loct"
                    is_loc = p.inflect({"loct"})
                    is_acc = p.inflect({"accs"})
                    word_curr = tokens[j].lower()
                    if is_loc and is_loc.word == word_curr:
                        return "loct"
                    if is_acc and is_acc.word == word_curr:
                        return "accs"
                if "VERB" in p.tag or "INFN" in p.tag:
                    break
            for j in range(idx + 1, min(len(tokens), idx + 4)):
                p = morph.parse(tokens[j])[0]
                word_norm = p.normal_form
                if word_norm in set(TIME_WORDS) | {
                    "январь",
                    "февраль",
                    "март",
                    "апрель",
                    "май",
                    "июнь",
                    "июль",
                    "август",
                    "сентябрь",
                    "октябрь",
                    "ноябрь",
                    "декабрь",
                }:
                    if is_integer_token(tokens[idx]):
                        parsed_num = parse_integer_token(tokens[idx])
                        if parsed_num is None:
                            return "loct"
                        _, unsigned = parsed_num
                        val = int(unsigned)
                        return "loct" if 1000 <= val <= 2100 else "accs"
                    return "loct"
            return "accs"
        if word_left == "по":
            if is_integer_token(tokens[idx]):
                parsed_num = parse_integer_token(tokens[idx])
                if parsed_num is None:
                    return "datv"
                _, unsigned = parsed_num
                val = int(unsigned)
                if val % 10 == 1 and val % 100 != 11:
                    return "datv"
                return "accs"
            return "datv"
        return prep_case

    blocked_by_noun = False
    for i in range(idx - 1, max(-1, idx - 3), -1):
        p = morph.parse(tokens[i])[0]
        if p.tag.POS == "NOUN":
            blocked_by_noun = True
            break

    for i in range(max(0, idx - 2), idx):
        p = morph.parse(tokens[i])[0]
        if blocked_by_noun and p.tag.POS in {"ADJF", "PRTF"}:
            continue
        if p.tag.POS in {"ADJF", "PRTF"} and p.tag.case:
            return p.tag.case

    for i in range(idx - 1, max(-1, idx - 5), -1):
        word_left = tokens[i].lower().strip(".,!?;:")
        p_verb = morph.parse(word_left)[0]
        if p_verb.normal_form in VERB_CASE:
            return VERB_CASE[p_verb.normal_form]
        if any(char in tokens[i] for char in ".!?"):
            break

    if idx < len(tokens) - 1:
        word_right = tokens[idx + 1].lower().strip(".,!?;:")
        if word_right:
            p_right = morph.parse(word_right)[0]
            noun_case = p_right.tag.case
            if not (
                is_case_reliable_noun(p_right)
                and (noun_case in {"gent", "datv", "ablt", "loct"} or "loc2" in p_right.tag)
            ):
                for entity_type, keywords in ENTITY_KEYWORDS.items():
                    if p_right.normal_form in keywords or word_right in keywords:
                        return ENTITY_DEFAULT_CASE[entity_type]

    if idx < len(tokens) - 1:
        word_right = tokens[idx + 1].lower().strip(".,!?;:")
        if word_right:
            p_right = morph.parse(word_right)[0]
            if is_case_reliable_noun(p_right):
                noun_case = p_right.tag.case
                if noun_case in {"datv", "ablt", "loct"} or "loc2" in p_right.tag:
                    return "loct" if noun_case == "loct" or "loc2" in p_right.tag else noun_case

    if idx == 0 or any(char in tokens[idx - 1] for char in ".!?"):
        return "nomn"
    return "nomn"


def detokenize(tokens: list[str]) -> str:
    parts: list[str] = []
    previous_was_newline = True
    for token in tokens:
        if not token:
            continue
        if "\n" in token or PARAGRAPH_BREAK_PLACEHOLDER in token:
            parts.append(token)
            previous_was_newline = True
            continue
        if parts and not previous_was_newline:
            parts.append(" ")
        parts.append(token)
        previous_was_newline = False
    text = "".join(parts)
    text = SENTENCE_PUNCTUATION_PATTERN.sub(r"\1", text)
    text = POINT_NUMBER_SPACING_PATTERN.sub("", text)
    text = POINT_WORD_PATTERN.sub(r"\1.", text)
    text = REPEATED_PUNCTUATION_PATTERN.sub(r"\1", text)
    text = DOUBLE_DOT_PATTERN.sub(".", text)
    text = COMMA_DOT_PATTERN.sub(".", text)
    text = DOT_COMMA_PATTERN.sub(".", text)
    text = OPEN_BRACKET_SPACING_PATTERN.sub(r"\1", text)
    text = CLOSE_BRACKET_SPACING_PATTERN.sub(r"\1", text)
    text = WORD_RANGE_HYPHEN_PATTERN.sub(" — ", text)
    return normalize_ascii_quote_pairs(text)



from __future__ import annotations

import re

import num2words

from .._morph import get_morph
from ..text_context import simple_tokenize
from ._constants import (
    DIGIT_PATTERN,
    NUMERIC_UNIT_RANGE_PATTERN,
    POST_NUMERAL_ABBREVIATION_PATTERNS,
    PREP_CASE,
    TIME_WORDS,
    UNITS_DATA,
)
from ._helpers import (
    build_number_token,
    detokenize,
    get_numeral_case,
    get_target_tags_for_number,
    inflect_numeral_string,
    inflect_unit_lemma,
    parse_integer_token,
    safe_inflect,
    should_consume_abbreviation_dot,
    should_keep_decimal_unit_dot,
)

GENITIVE_RANGE_CONTEXT_STEMS = (
    "диаметр",
    "объем",
    "объём",
    "ширин",
    "высот",
    "длин",
    "глубин",
    "радиус",
    "толщин",
    "размер",
)


def _resolve_range_unit_info(unit_raw: str):
    unit_lower = unit_raw.lower().strip(".")
    direct = UNITS_DATA.get(unit_lower)
    if direct is not None:
        return direct

    parts = [part for part in unit_lower.split() if part]
    if len(parts) < 2:
        return None

    morph = get_morph()
    noun_parse = next(
        (
            candidate
            for candidate in morph.parse(parts[-1])
            if "NOUN" in candidate.tag
        ),
        None,
    )
    if noun_parse is None:
        return None

    if len(parts) == 2:
        head_parse = next(
            (
                candidate
                for candidate in morph.parse(parts[0])
                if candidate.tag.POS in {"ADJF", "PRTF"}
            ),
            None,
        )
        if head_parse is None:
            return None
        lemma = f"{head_parse.normal_form} {noun_parse.normal_form}"
    else:
        first_parse = next(
            (
                candidate
                for candidate in morph.parse(parts[0])
                if candidate.tag.POS in {"ADJF", "PRTF"}
            ),
            None,
        )
        second_parse = next(
            (
                candidate
                for candidate in morph.parse(parts[1])
                if candidate.tag.POS in {"ADJF", "PRTF"}
            ),
            None,
        )
        if first_parse is None or second_parse is None:
            return None
        lemma = (
            f"{first_parse.normal_form} {second_parse.normal_form}"
            f" {noun_parse.normal_form}"
        )

    for info in UNITS_DATA.values():
        if info[0] == lemma:
            return info
    return None


def normalize_cardinal_numerals(text: str) -> str:
    morph = get_morph()
    tokens = simple_tokenize(text)
    result_tokens: list[str] = []

    def is_word_token(token: str) -> bool:
        return bool(re.fullmatch(r"[^\W\d_]+", token, re.UNICODE))

    def is_ambiguous_preposition_token(token: str) -> bool:
        clean = token.lower().strip('.,:;!"«»()[]{}')
        return bool(clean) and " " not in clean and clean in PREP_CASE

    def should_skip_unit_candidate(start_index: int, token_span: int, unit_raw: str) -> bool:
        chunk = tokens[start_index : start_index + token_span]
        if (
            token_span == 1
            and unit_raw == unit_raw.strip('.,:;!"«»()[]{}')
            and is_ambiguous_preposition_token(unit_raw)
            and start_index + token_span < len(tokens)
        ):
            next_token = tokens[start_index + token_span].strip('.,:;!"«»()[]{}')
            if next_token:
                return True
        return (
            token_span > 1
            and all(is_word_token(token) for token in chunk)
            and any(is_ambiguous_preposition_token(token) for token in chunk[1:])
        )

    def extract_unit_candidate(start_index: int) -> tuple[str, str, int] | None:
        best_match: tuple[str, str, int] | None = None
        max_end = min(len(tokens), start_index + 4)
        for end_index in range(start_index + 1, max_end + 1):
            chunk = tokens[start_index:end_index]
            if any("\n" in token for token in chunk):
                break
            candidate_raw = "".join(chunk)
            candidate_key = candidate_raw.lower().strip(".")
            if candidate_key in UNITS_DATA:
                if should_skip_unit_candidate(
                    start_index, end_index - start_index, candidate_raw
                ):
                    continue
                best_match = (candidate_key, candidate_raw, end_index - start_index)
        return best_match

    def is_redundant_unit_token(token_index: int, expected_lemma: str) -> bool:
        if token_index >= len(tokens):
            return False
        token_lower = tokens[token_index].lower().strip('.,:;!"«»()[]{}')
        if not token_lower:
            return False
        parsed = morph.parse(token_lower)
        if not parsed:
            return False
        candidate = parsed[0]
        return "NOUN" in candidate.tag and candidate.normal_form == expected_lemma

    i = 0
    while i < len(tokens):
        token = tokens[i]
        parsed_num = parse_integer_token(token)
        if parsed_num is None:
            result_tokens.append(token)
            i += 1
            continue

        is_negative, clean_token = parsed_num
        val = int(clean_token)
        case = get_numeral_case(tokens, i)
        inflected_num = inflect_numeral_string(clean_token, case)
        num_words = build_number_token(token, clean_token, inflected_num, is_negative)

        if i + 2 < len(tokens):
            adj_token = tokens[i + 1]
            noun_token = tokens[i + 2]
            clean_adj = adj_token.lower().strip('.,:;!"«»()[]{}')
            clean_noun = noun_token.lower().strip('.,:;!"«»()[]{}')
            p_adj = morph.parse(clean_adj)[0]
            p_noun = morph.parse(clean_noun)[0]
            if ("ADJF" in p_adj.tag or "PRTF" in p_adj.tag) and "NOUN" in p_noun.tag:
                gender = p_noun.tag.gender
                is_anim = "anim" in p_noun.tag
                target_num_case = case
                if case == "accs":
                    rem100 = val % 100
                    rem10 = val % 10
                    if rem10 in (2, 3, 4) and rem100 not in (12, 13, 14):
                        target_num_case = "gent" if is_anim else "nomn"
                num_words = build_number_token(
                    token,
                    clean_token,
                    inflect_numeral_string(clean_token, target_num_case, gender),
                    is_negative,
                )
                result_tokens.extend([num_words, adj_token, noun_token])
                i += 3
                continue

        if i + 1 < len(tokens):
            noun_token = tokens[i + 1]
            next_token_lower = noun_token.lower().strip('.,:;!"«»()[]{}')
            unit_candidate = extract_unit_candidate(i + 1)
            unit_token_span = 1
            unit_raw = noun_token
            if next_token_lower == "°" and i + 2 < len(tokens):
                degree_suffix = tokens[i + 2].lower().strip('.,:;!"«»()[]{}')
                if degree_suffix in {"c", "k", "f"}:
                    next_token_lower = f"°{degree_suffix}"
                    noun_token = f"{noun_token}{tokens[i + 2]}"
                    unit_raw = noun_token
                    unit_token_span = 2
            elif unit_candidate:
                next_token_lower, unit_raw, unit_token_span = unit_candidate
            if should_skip_unit_candidate(i + 1, unit_token_span, unit_raw):
                unit_info = None
            else:
                unit_info = UNITS_DATA.get(next_token_lower)
            if unit_info:
                preserve_unit_dot = (
                    unit_raw.endswith(".")
                    and not should_consume_abbreviation_dot(
                        tokens, i + unit_token_span
                    )
                )
                lemma, u_gender, u_category, *u_suffix = unit_info
                multipliers = {"тысяча", "миллион", "миллиард", "триллион"}
                currency_symbol_units = {
                    "$",
                    "€",
                    "₽",
                    "£",
                    "¥",
                    "₴",
                    "₸",
                    "₺",
                    "₹",
                    "¢",
                    "₪",
                    "₩",
                    "₫",
                    "₱",
                    "₦",
                }
                if (
                    u_category == "money"
                    and next_token_lower in currency_symbol_units
                    and i + 2 < len(tokens)
                ):
                    multiplier_token = tokens[i + 2]
                    multiplier_lower = multiplier_token.lower().strip('.,:;!"«»()[]{}')
                    p_multiplier = morph.parse(multiplier_lower)[0]
                    if (
                        "NOUN" in p_multiplier.tag
                        and p_multiplier.normal_form in multipliers
                    ):
                        target_num_case = case
                        if case == "accs":
                            rem100 = val % 100
                            rem10 = val % 10
                            if rem10 in (2, 3, 4) and rem100 not in (12, 13, 14):
                                target_num_case = "nomn"
                        multiplier_gender = p_multiplier.tag.gender or "masc"
                        num_words = build_number_token(
                            token,
                            clean_token,
                            inflect_numeral_string(
                                clean_token, target_num_case, multiplier_gender
                            ),
                            is_negative,
                        )
                        result_tokens.extend(
                            [
                                num_words,
                                multiplier_token,
                                inflect_unit_lemma(lemma, {"gent", "plur"}),
                            ]
                        )
                        i += 4 if is_redundant_unit_token(i + 3, lemma) else 3
                        continue

                target_num_case = case
                if case == "accs":
                    rem100 = val % 100
                    rem10 = val % 10
                    if rem10 in (2, 3, 4) and rem100 not in (12, 13, 14):
                        target_num_case = "nomn"
                num_words = build_number_token(
                    token,
                    clean_token,
                    inflect_numeral_string(clean_token, target_num_case, u_gender),
                    is_negative,
                )
                inflected_unit = inflect_unit_lemma(
                    lemma, get_target_tags_for_number(val, case, u_gender)
                )
                full_unit = inflected_unit + (f" {u_suffix[0]}" if u_suffix else "")
                match_unit = (
                    re.search(re.escape(next_token_lower), noun_token, re.IGNORECASE)
                    if unit_token_span == 1
                    else None
                )
                if match_unit:
                    full_unit = (
                        noun_token[: match_unit.start()]
                        + full_unit
                        + noun_token[match_unit.end() :]
                    )
                if preserve_unit_dot and should_keep_decimal_unit_dot(
                    detokenize(tokens[i + 1 + unit_token_span :])
                ):
                    full_unit += "."
                result_tokens.extend([num_words, full_unit])
                step = 1 + unit_token_span
                if (
                    unit_token_span == 1
                    and next_token_lower.startswith("°")
                    and len(next_token_lower) == 2
                ):
                    step = 3
                if i + step < len(tokens) and should_consume_abbreviation_dot(tokens, i + step):
                    step += 1
                if is_redundant_unit_token(i + step, lemma):
                    step += 1
                if lemma in multipliers and i + 2 < len(tokens):
                    next_next_token = tokens[i + 2]
                    nn_token_lower = next_next_token.lower().strip('.,:;!"«»()[]{}')
                    nn_unit_info = UNITS_DATA.get(nn_token_lower)
                    if nn_unit_info:
                        nn_lemma, _, _, *nn_suffix = nn_unit_info
                        p_nn = morph.parse(nn_lemma)[0]
                        nn_inflected = safe_inflect(p_nn, {"gent", "plur"})
                        if nn_suffix:
                            nn_inflected += f" {nn_suffix[0]}"
                        result_tokens.append(nn_inflected)
                        i += step + 1
                        continue
                i += step
                continue

            p_noun = morph.parse(next_token_lower)[0]
            if "NOUN" in p_noun.tag:
                gender = TIME_WORDS.get(next_token_lower, p_noun.tag.gender)
                is_anim = "anim" in p_noun.tag
                target_num_case = case
                if case == "accs":
                    rem100 = val % 100
                    rem10 = val % 10
                    if rem10 in (2, 3, 4) and rem100 not in (12, 13, 14):
                        target_num_case = "gent" if is_anim else "nomn"
                num_words = build_number_token(
                    token,
                    clean_token,
                    inflect_numeral_string(clean_token, target_num_case, gender),
                    is_negative,
                )
                result_tokens.extend([num_words, noun_token])
                i += 2
                continue

        result_tokens.append(num_words)
        i += 1
    return detokenize(result_tokens)


def normalize_numeric_unit_ranges(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        left = match.group("left")
        right = match.group("right")
        unit_raw = match.group("unit")
        unit_info = _resolve_range_unit_info(unit_raw)
        if not unit_info:
            return match.group(0)
        context = text[max(0, match.start() - 40) : match.start()]
        tokens_left = simple_tokenize(context)
        case = get_numeral_case(tokens_left + [left], len(tokens_left))
        normalized_context = " ".join(tokens_left[-6:]).lower()
        if any(stem in normalized_context for stem in GENITIVE_RANGE_CONTEXT_STEMS):
            case = "gent"
        lemma, u_gender, _, *u_suffix = unit_info

        def numeral_for_range(value_text: str) -> str:
            value = int(value_text)
            target_case = case
            if case == "accs":
                rem100 = value % 100
                rem10 = value % 10
                if rem10 in (2, 3, 4) and rem100 not in (12, 13, 14):
                    target_case = "nomn"
            return inflect_numeral_string(value_text, target_case, u_gender)

        right_value = int(right)
        unit_words = inflect_unit_lemma(
            lemma, get_target_tags_for_number(right_value, case, u_gender)
        )
        if u_suffix:
            unit_words += f" {u_suffix[0]}"
        return f"{numeral_for_range(left)} — {numeral_for_range(right)} {unit_words}"

    return NUMERIC_UNIT_RANGE_PATTERN.sub(repl, text)


def normalize_remaining_post_numeral_abbreviations(text: str) -> str:
    def preserve_sentence_boundary(
        match: re.Match[str], replacement: str, source_text: str
    ) -> str:
        tail = source_text[match.end() :]
        stripped_tail = tail.lstrip()
        if not stripped_tail:
            return f"{replacement}."
        next_char = stripped_tail[0]
        if next_char in "\n.!?…":
            return f"{replacement}."
        return replacement

    for pattern, replacement in POST_NUMERAL_ABBREVIATION_PATTERNS:
        text = pattern.sub(
            lambda match, repl=replacement, src=text: preserve_sentence_boundary(
                match, repl, src
            ),
            text,
        )
    return text


def normalize_all_digits_everywhere(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        try:
            return num2words.num2words(int(match.group(0)), lang="ru")
        except Exception:
            return match.group(0)

    return DIGIT_PATTERN.sub(repl, text)



from __future__ import annotations

import logging
import pickle
import re
import threading
from pathlib import Path
from typing import Any

logger = logging.getLogger(__name__)
_DICTIONARY_CACHE_VERSION = "v2"


def _compile_simple_mapping_patterns(
    mapping: dict[str, str],
) -> list[tuple[re.Pattern[str], str]]:
    patterns: list[tuple[re.Pattern[str], str]] = []
    for key in sorted(mapping.keys(), key=len, reverse=True):
        patterns.append(
            (
                re.compile(
                    rf"(?<!\w){re.escape(key)}(?!\w)",
                    re.UNICODE | re.IGNORECASE,
                ),
                mapping[key],
            )
        )
    return patterns


class DictionaryNormalizer:
    """Apply .dic substitution rules in filename order."""

    def __init__(
        self,
        dictionaries_path: str | Path | None = None,
        exclude_files: list[str] | None = None,
        include_only_files: list[str] | None = None,
    ) -> None:
        self.dictionaries_path = (
            Path(dictionaries_path)
            if dictionaries_path
            else Path(__file__).resolve().parent / "dictionaries"
        )
        self.exclude_files = set(exclude_files or [])
        self.include_only_files = set(include_only_files or [])
        self._file_rules: list[tuple[str, Any]] = []
        self._runtime_file_rules: list[tuple[str, Any]] = []
        self._total_dic = 0

        if not self._try_load_from_cache():
            self._load_all_dictionaries()
            self._save_to_cache()
        self._build_runtime_rules()

    def _get_cache_path(self) -> Path:
        base = f"dictionaries_{_DICTIONARY_CACHE_VERSION}"
        if self.include_only_files:
            import hashlib

            digest = hashlib.md5(
                ",".join(sorted(self.include_only_files)).encode()
            ).hexdigest()[:8]
            base += f"_include_{digest}"
        elif self.exclude_files:
            import hashlib

            digest = hashlib.md5(
                ",".join(sorted(self.exclude_files)).encode()
            ).hexdigest()[:8]
            base += f"_exclude_{digest}"
        return self.dictionaries_path / f"{base}.pkl"

    def _try_load_from_cache(self) -> bool:
        cache_path = self._get_cache_path()
        if not cache_path.exists():
            return False
        try:
            cache_mtime = cache_path.stat().st_mtime
            for dic_file in self.dictionaries_path.rglob("*.dic"):
                if dic_file.stat().st_mtime > cache_mtime:
                    return False
            with cache_path.open("rb") as fh:
                data = pickle.load(fh)
            if data.get("cache_version") != _DICTIONARY_CACHE_VERSION:
                return False
            if set(data.get("exclude_files", [])) != self.exclude_files:
                return False
            if set(data.get("include_only_files", [])) != self.include_only_files:
                return False
            self._file_rules = data["rules"]
            self._total_dic = data["dic_count"]
            return True
        except Exception as exc:
            logger.warning("Failed to load dictionary cache %s: %s", cache_path, exc)
            return False

    def _save_to_cache(self) -> None:
        try:
            with self._get_cache_path().open("wb") as fh:
                pickle.dump(
                    {
                        "rules": self._file_rules,
                        "dic_count": self._total_dic,
                        "cache_version": _DICTIONARY_CACHE_VERSION,
                        "exclude_files": list(self.exclude_files),
                        "include_only_files": list(self.include_only_files),
                    },
                    fh,
                )
        except Exception as exc:
            logger.warning("Failed to save dictionary cache: %s", exc)

    def _load_all_dictionaries(self) -> None:
        if not self.dictionaries_path.exists():
            return
        files = sorted(self.dictionaries_path.rglob("*.dic"))
        for filepath in files:
            relative_name = filepath.relative_to(self.dictionaries_path).as_posix()
            if not self.include_only_files and relative_name.startswith("latinization/"):
                continue
            if self.include_only_files and (
                filepath.name not in self.include_only_files
                and relative_name not in self.include_only_files
            ):
                continue
            if filepath.name in self.exclude_files or relative_name in self.exclude_files:
                continue
            try:
                chunks = self._load_dic_file(filepath)
                if chunks:
                    self._file_rules.append(("dic", chunks))
                    self._total_dic += sum(
                        len(chunk[1]) if chunk[0] != "regex_batch" else len(chunk[1][1])
                        for chunk in chunks
                    )
            except Exception as exc:
                logger.error("Error loading %s: %s", filepath.name, exc)

    def _load_dic_file(self, path: Path) -> list[tuple[str, Any]]:
        chunks: list[tuple[str, Any]] = []
        current_simple: dict[str, str] = {}
        current_regex: list[tuple[re.Pattern[str], str]] = []

        def flush() -> None:
            nonlocal current_simple, current_regex
            if current_simple:
                if len(current_simple) < 1000:
                    try:
                        keys = sorted(current_simple.keys(), key=len, reverse=True)
                        pattern = re.compile(
                            r"(?<!\w)(?:" + "|".join(map(re.escape, keys)) + r")(?!\w)",
                            re.UNICODE | re.IGNORECASE,
                        )
                        chunks.append(("regex_batch", (pattern, current_simple)))
                    except Exception:
                        chunks.append(("simple", current_simple))
                else:
                    chunks.append(("simple", current_simple))
                current_simple = {}
            if current_regex:
                chunks.append(("regex", current_regex))
                current_regex = []

        with path.open("r", encoding="utf-8", errors="replace") as fh:
            for line in fh:
                line = line.rstrip("\n\r")
                if not line or line.startswith("#"):
                    continue
                idx = line.find("=")
                if idx == -1:
                    continue
                source = line[:idx]
                target = line[idx + 1 :]
                if not source:
                    continue

                is_simple = not source.startswith("$") and not any(
                    char in r"*\[](){}^$.|+?" for char in source
                )
                if is_simple:
                    if current_regex:
                        flush()
                    current_simple[source.lower()] = target
                    continue

                if current_simple:
                    flush()
                try:
                    anchor_start = source.startswith("$")
                    if anchor_start:
                        source = source[1:]
                    if not source:
                        continue
                    pattern_str, replacement = self._dic_pattern_to_regex(
                        source, target, anchor_start
                    )
                    current_regex.append(
                        (
                            re.compile(pattern_str, re.UNICODE | re.IGNORECASE),
                            replacement,
                        )
                    )
                except re.error:
                    continue

        flush()
        return chunks

    def _build_runtime_rules(self) -> None:
        runtime_rules: list[tuple[str, Any]] = []
        for file_type, rules in self._file_rules:
            if file_type != "dic":
                runtime_rules.append((file_type, rules))
                continue
            runtime_chunks: list[tuple[str, Any]] = []
            for chunk_type, chunk_data in rules:
                if chunk_type == "simple":
                    runtime_chunks.append(
                        ("simple_compiled", _compile_simple_mapping_patterns(chunk_data))
                    )
                else:
                    runtime_chunks.append((chunk_type, chunk_data))
            runtime_rules.append((file_type, runtime_chunks))
        self._runtime_file_rules = runtime_rules

    def _dic_pattern_to_regex(
        self, source: str, target: str, anchor_start: bool = False
    ) -> tuple[str, str]:
        # Treat *foo* as an in-word substring replacement so regex.sub can
        # replace every occurrence across the text, not just one fragment
        # inside a whitespace token.
        if source.startswith("*") and source.endswith("*") and source.count("*") == 2:
            pattern = re.escape(source[1:-1])
            if anchor_start:
                pattern = "^" + pattern
            return pattern, target

        parts = source.split("*")
        regex_parts: list[str] = []
        replacement_parts = [target]
        total_groups = 0

        if parts[0] == "":
            regex_parts.append(r"(\S*)")
            total_groups += 1
            replacement_parts.insert(0, r"\g<1>")
            parts.pop(0)
        elif parts[0] and parts[0][0].isalnum():
            regex_parts.append(r"(?<!\w)")

        has_suffix_wildcard = False
        if parts and parts[-1] == "":
            has_suffix_wildcard = True
            parts.pop()

        for index, part in enumerate(parts):
            if index > 0:
                regex_parts.append(r"\S*")
            if part:
                regex_parts.append(re.escape(part))

        if has_suffix_wildcard:
            regex_parts.append(r"(\S*)")
            total_groups += 1
            replacement_parts.append(f"\\g<{total_groups}>")
        elif parts and parts[-1] and parts[-1][-1].isalnum():
            regex_parts.append(r"(?!\w)")

        pattern = "".join(regex_parts)
        if anchor_start:
            pattern = "^" + pattern
        return pattern, "".join(replacement_parts)

    def _apply_simple_chunk(
        self, text: str, patterns: list[tuple[re.Pattern[str], str]]
    ) -> str:
        for pattern, replacement in patterns:
            text = pattern.sub(replacement, text)
        return text

    def _apply_dic_rules(self, text: str, chunks: list[tuple[str, Any]]) -> str:
        for chunk_type, chunk_data in chunks:
            if chunk_type == "simple_compiled":
                text = self._apply_simple_chunk(text, chunk_data)
            elif chunk_type == "regex_batch":
                pattern, mapping = chunk_data
                text = pattern.sub(
                    lambda match: mapping.get(match.group().lower(), match.group()),
                    text,
                )
            else:
                for pattern, replacement in chunk_data:
                    try:
                        text = pattern.sub(replacement, text)
                    except Exception:
                        continue
        return text

    def apply(self, text: str, *, strip_unmatched_latin: bool = False) -> str:
        if not self._runtime_file_rules:
            return text
        for file_type, rules in self._runtime_file_rules:
            if file_type == "dic":
                text = self._apply_dic_rules(text, rules)
        if strip_unmatched_latin:
            return re.sub(r"[A-Za-z]", "", text)
        return text

    def get_rule_count(self) -> int:
        return self._total_dic


_normalizers: dict[
    tuple[str, tuple[str, ...], tuple[str, ...]], DictionaryNormalizer
] = {}
_normalizers_lock = threading.Lock()


def get_dictionary_normalizer(
    dictionaries_path: str | Path | None = None,
    *,
    exclude_files: list[str] | None = None,
    include_only_files: list[str] | None = None,
) -> DictionaryNormalizer:
    key = (
        str(
            Path(dictionaries_path)
            if dictionaries_path
            else Path(__file__).resolve().parent / "dictionaries"
        ),
        tuple(sorted(exclude_files or [])),
        tuple(sorted(include_only_files or [])),
    )
    if key not in _normalizers:
        with _normalizers_lock:
            if key not in _normalizers:
                _normalizers[key] = DictionaryNormalizer(
                    dictionaries_path=key[0],
                    exclude_files=list(key[1]),
                    include_only_files=list(key[2]),
                )
    return _normalizers[key]


def apply_dictionary_rules(
    text: str,
    *,
    enabled: bool,
    dictionaries_path: str | Path | None = None,
    include_only_files: tuple[str, ...] = (),
    exclude_files: tuple[str, ...] = (),
    strip_unmatched_latin: bool = False,
) -> str:
    if not enabled:
        return text
    normalizer = get_dictionary_normalizer(
        dictionaries_path=dictionaries_path,
        include_only_files=list(include_only_files),
        exclude_files=list(exclude_files),
    )
    return normalizer.apply(text, strip_unmatched_latin=strip_unmatched_latin)



from __future__ import annotations

import re

import num2words

from .._morph import get_morph
from ._constants import FRACTION_PATTERN

GENITIVE_FRACTION_CONTEXT_PATTERN = re.compile(r"\b(с|со|от|до|из|без|у)\s+$")
DATIVE_FRACTION_CONTEXT_PATTERN = re.compile(r"\b(к|по)\s+$")
ACCUSATIVE_FRACTION_CONTEXT_PATTERN = re.compile(r"\b(в|на|через)\s+$")
PREPOSITIONAL_FRACTION_CONTEXT_PATTERN = re.compile(r"\b(о|об|при)\s+$")


def normalize_fractions(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        num = int(match.group(1))
        denom = int(match.group(2))
        context = text[max(0, match.start() - 10) : match.start()].lower()
        case = "nomn"
        if GENITIVE_FRACTION_CONTEXT_PATTERN.search(context):
            case = "gent"
        elif DATIVE_FRACTION_CONTEXT_PATTERN.search(context):
            case = "datv"
        elif ACCUSATIVE_FRACTION_CONTEXT_PATTERN.search(context):
            case = "accs"
        elif PREPOSITIONAL_FRACTION_CONTEXT_PATTERN.search(context):
            case = "loct"
        try:
            num_text = num2words.num2words(num, lang="ru")
        except Exception:
            return match.group(0)
        morph = get_morph()
        if case != "nomn":
            num_text = " ".join(
                (p.inflect({case}).word if p.inflect({case}) else part)
                for part in num_text.split()
                for p in [morph.parse(part)[0]]
            )
        last_num_word = num_text.split()[-1]
        p_last = morph.parse(last_num_word)[0]
        if num % 10 == 1 and num % 100 != 11:
            inf = p_last.inflect({case, "femn", "sing"})
            if inf:
                arr = num_text.split()
                arr[-1] = inf.word
                num_text = " ".join(arr)
        elif num % 10 == 2 and num % 100 != 12 and case in ["nomn", "accs"]:
            inf = p_last.inflect({case, "femn"})
            if inf:
                arr = num_text.split()
                arr[-1] = inf.word
                num_text = " ".join(arr)
        try:
            denom_text = num2words.num2words(denom, lang="ru", to="ordinal")
        except Exception:
            return match.group(0)
        words = denom_text.split()
        p = morph.parse(words[-1])[0]
        is_sing_1 = num % 10 == 1 and num % 100 != 11
        inflected = p.inflect(
            {case, "femn", "sing"}
            if is_sing_1
            else ({"gent", "plur"} if case in ["nomn", "accs"] else {case, "plur"})
        )
        if inflected:
            words[-1] = inflected.word
        return f"{num_text} {' '.join(words)}"

    return FRACTION_PATTERN.sub(repl, text)



from __future__ import annotations

import re

from ._morph import get_morph
from .numerals._constants import ALL_UNITS, CURRENCY_STANDALONE, ENTITY_KEYWORDS
from .text_context import normalize_context_token, simple_tokenize

YEAR_MIN = 1000
YEAR_MAX = 2100
YEAR_THOUSANDS_SEPARATORS = " \u00A0\u2009\u202F"
SPACED_THOUSANDS_TAIL = rf"(?![{YEAR_THOUSANDS_SEPARATORS}]\d{{3}}\b)"
YEAR_ANY_NUMBER_PATTERN = rf"\d+{SPACED_THOUSANDS_TAIL}"
YEAR_RANGE_NUMBER_PATTERN = rf"\d{{3,4}}{SPACED_THOUSANDS_TAIL}"
YEAR_IMPLICIT_PREP_PATTERN = rf"(?:1\d|20)\d{{2}}{SPACED_THOUSANDS_TAIL}"

_TRIVIA_TOKENS = {",", ";", ":", ".", "!", "?", "…", "»", '"', "”", ")", "]", "}"}
_NON_YEAR_EXCEPTION_LEMMAS = {"раз", "место"}
_NON_YEAR_LEXICON = (
    ALL_UNITS
    | set(CURRENCY_STANDALONE)
    | ENTITY_KEYWORDS["money"]
    | ENTITY_KEYWORDS["percent"]
    | ENTITY_KEYWORDS["measure"]
)


def _leading_context_tokens(text: str, start: int, limit: int = 3) -> list[str]:
    tokens: list[str] = []
    for token in simple_tokenize(text[start:]):
        normalized = normalize_context_token(token)
        candidate = normalized or token.strip()
        if not candidate or candidate in _TRIVIA_TOKENS:
            continue
        tokens.append(candidate.lower())
        if len(tokens) >= limit:
            break
    return tokens


def _consume_following_number(tokens: list[str], start: int) -> tuple[str | None, int]:
    if start >= len(tokens) or not tokens[start].isdigit():
        return None, start
    parts = [tokens[start]]
    idx = start + 1
    while idx < len(tokens) and tokens[idx].isdigit() and len(tokens[idx]) == 3:
        parts.append(tokens[idx])
        idx += 1
    return "".join(parts), idx


def is_plausible_year(value: int) -> bool:
    return YEAR_MIN <= value <= YEAR_MAX


def _is_non_year_following_token(token: str) -> bool:
    if token in _NON_YEAR_LEXICON:
        return True
    parsed = get_morph().parse(token)
    if not parsed:
        return False
    lemma = parsed[0].normal_form
    return lemma in _NON_YEAR_LEXICON or lemma in _NON_YEAR_EXCEPTION_LEMMAS


def should_treat_as_implicit_year(
    text: str,
    start: int,
    *,
    explicit_year_word_pattern: re.Pattern[str],
    year_suffix_tail_pattern: re.Pattern[str] | None = None,
) -> bool:
    if year_suffix_tail_pattern and year_suffix_tail_pattern.match(text, start):
        return False
    if explicit_year_word_pattern.match(text, start):
        return False

    following_tokens = _leading_context_tokens(text, start, limit=4)
    if not following_tokens:
        return True
    if len(following_tokens) >= 2 and following_tokens[0] in {"до", "по"}:
        range_number, next_idx = _consume_following_number(following_tokens, 1)
        if range_number is not None:
            if 3 <= len(range_number) <= 4:
                return False
            if len(range_number) > 4:
                return False
            if next_idx < len(following_tokens) and _is_non_year_following_token(
                following_tokens[next_idx]
            ):
                return False
    return not _is_non_year_following_token(following_tokens[0])



from __future__ import annotations

import re

from ._constants import (
    CURRENCY_STANDALONE,
    GREEK_LETTERS,
    MATH_SYMBOLS,
    STANDALONE_CURRENCY_PATTERN,
)

MATH_EXPRESSION_CHAR_PATTERN = re.compile(r"[0-9A-Za-zА-Яа-яЁё_,.%()+\-/*^°№]")
APPROXIMATE_NUMBER_PATTERN = re.compile(r"(?<!\w)~\s*(?=(?:\ue001)?\d)")


def _extract_math_expression_side(text: str, start: int, step: int) -> str:
    chars: list[str] = []
    idx = start
    saw_non_space = False
    while 0 <= idx < len(text):
        char = text[idx]
        if char.isspace() or MATH_EXPRESSION_CHAR_PATTERN.fullmatch(char):
            chars.append(char)
            if not char.isspace():
                saw_non_space = True
            idx += step
            continue
        break
    if not saw_non_space:
        return ""
    side = "".join(reversed(chars)) if step < 0 else "".join(chars)
    return side.strip()


def _contains_digit(fragment: str) -> bool:
    return any(char.isdigit() for char in fragment)


def _should_read_equals_as_ravno(text: str, idx: int) -> bool:
    prev_char = text[idx - 1] if idx > 0 else ""
    next_char = text[idx + 1] if idx + 1 < len(text) else ""
    if prev_char in "<>!=" or next_char in "<>!=":
        return False

    left = _extract_math_expression_side(text, idx - 1, -1)
    right = _extract_math_expression_side(text, idx + 1, 1)
    if not left or not right:
        return False
    return _contains_digit(left) or _contains_digit(right)


def _normalize_contextual_equals(text: str) -> str:
    if "=" not in text:
        return text

    parts: list[str] = []
    last_index = 0
    for match in re.finditer(r"=", text):
        idx = match.start()
        if not _should_read_equals_as_ravno(text, idx):
            continue
        parts.append(text[last_index:idx])
        parts.append(" равно ")
        last_index = idx + 1

    if last_index == 0:
        return text

    parts.append(text[last_index:])
    return "".join(parts)


def normalize_greek_letters(text: str) -> str:
    greek_chars = "".join(re.escape(char) for char in GREEK_LETTERS)
    pattern = re.compile(rf"[{greek_chars}]+")

    def repl(match: re.Match[str]) -> str:
        token = match.group(0)
        if len(token) != 1:
            return token
        return GREEK_LETTERS.get(token, token)

    return pattern.sub(repl, text)


def normalize_math_symbols(text: str) -> str:
    text = _normalize_contextual_equals(text)
    text = APPROXIMATE_NUMBER_PATTERN.sub("примерно ", text)
    for char, replacement in MATH_SYMBOLS.items():
        text = text.replace(char, replacement)
    return text


def normalize_standalone_currency(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        return CURRENCY_STANDALONE[match.group(0)]

    return STANDALONE_CURRENCY_PATTERN.sub(repl, text)



"""Russian text normalization library."""

try:
    from .options import NormalizeOptions
    from .pipeline import Normalizer, normalize, preprocess_text
except ImportError:  # pragma: no cover - supports direct file import during test collection
    from ru_normalizr.options import NormalizeOptions
    from ru_normalizr.pipeline import Normalizer, normalize, preprocess_text

__all__ = ["NormalizeOptions", "Normalizer", "normalize", "preprocess_text"]
__version__ = "0.2.0"



from __future__ import annotations

import re

import num2words
import roman

from .options import NormalizeOptions
from .preprocess_utils import classify_bracketed_numeric_content


def num_to_word(value: int) -> str:
    try:
        return num2words.num2words(value, lang="ru")
    except Exception:
        return str(value)


def roman_to_word(roman_str: str) -> str:
    try:
        return num_to_word(roman.fromRoman(roman_str.upper()))
    except roman.InvalidRomanNumeralError:
        return roman_str


def convert_numeric_sequence(sequence: str) -> str:
    clean_seq = sequence.rstrip(".")
    if re.match(r"^[IVXLCDM]+$", clean_seq, re.IGNORECASE):
        return roman_to_word(clean_seq)
    parts = [part for part in clean_seq.split(".") if part]
    words: list[str] = []
    for part in parts:
        try:
            words.append(num_to_word(int(part)))
        except ValueError:
            words.append(part)
    return " точка ".join(words)


def convert_line_numbering(text: str) -> str:
    def convert_numbering(match: re.Match[str]) -> str:
        converted = convert_numeric_sequence(match.group(2)).capitalize()
        return f"{match.group(1)}{converted}.{match.group(3)}"

    pattern = r"^([ \t]*)([IVXLCDM]+\.|[\d]+(?:\.[\d]+)*\.)([ \t]*)"
    return "\n".join(
        re.sub(pattern, convert_numbering, line, flags=re.IGNORECASE)
        for line in text.split("\n")
    )


def convert_bracketed_numbers(
    text: str, options: NormalizeOptions | None = None
) -> str:
    active = options or NormalizeOptions()
    pattern = r"([(\[{])\s*([^()\[\]{}]+?)\*?\s*([)\]}])"
    pairs = {"(": ")", "[": "]", "{": "}"}

    def repl(match: re.Match[str]) -> str:
        opener, sequence, closer = match.group(1), match.group(2), match.group(3)
        if pairs.get(opener) != closer:
            return match.group(0)

        stripped = sequence.strip()
        if re.match(r"^[IVXLCDM]+$", stripped, re.IGNORECASE):
            kind = "reference"
        else:
            kind = classify_bracketed_numeric_content(stripped)
        if kind != "reference":
            return match.group(0)

        should_remove = False
        if active.remove_links:
            should_remove = True
            try:
                if stripped.isdigit():
                    value = int(stripped)
                    min_val, max_val = active.remove_links_ignore_interval
                    if min_val <= value <= max_val:
                        should_remove = False
            except Exception:
                pass
        if should_remove:
            return ""
        return match.group(0)

    return re.sub(pattern, repl, text, flags=re.IGNORECASE)



from __future__ import annotations

import re

import roman

from ._morph import get_morph
from .constants import KNOWN_ABBREVIATIONS
from .numerals._helpers import get_numeral_case
from .numerals._hyphen import is_safe_numeric_hyphen_unit
from .options import NormalizeOptions
from .ordinal_utils import (
    choose_noun_parse,
    noun_parse_case,
    render_ordinal,
    render_ordinal_from_noun_word,
)
from .text_context import normalize_context_token, simple_tokenize

_ROMAN_ABBREVIATION_EXCEPTIONS = frozenset(
    {
        *(abbr for abbr in KNOWN_ABBREVIATIONS if re.fullmatch(r"[A-Z-]+", abbr)),
        "DC",
        "MD",
        "CV",
        "CI",
        "DI",
        "MC",
        "CM",
        "MM",
    }
)

_REGNAL_CASE_CONTEXT = {
    "в": "loct",
    "во": "loct",
    "на": "loct",
    "о": "loct",
    "об": "loct",
    "обо": "loct",
    "при": "loct",
    "к": "datv",
    "ко": "datv",
    "по": "datv",
    "с": "gent",
    "со": "gent",
    "из": "gent",
    "от": "gent",
    "до": "gent",
    "у": "gent",
    "без": "gent",
    "для": "gent",
    "около": "gent",
    "после": "gent",
    "ради": "gent",
    "между": "ablt",
    "над": "ablt",
    "под": "ablt",
    "перед": "ablt",
    "пред": "ablt",
}
_ROMAN_CONTEXT_LEMMAS = frozenset(
    {
        "век",
        "столетие",
        "тысячелетие",
        "глава",
        "раздел",
        "часть",
        "том",
        "книга",
        "квартал",
    }
)
_ROMAN_HEADING_LEMMAS = frozenset(
    {"глава", "раздел", "часть", "том", "книга", "квартал"}
)
_ROMAN_ABBREVIATION_TO_CONTEXT = {
    "в": ("век", "sing"),
    "вв": ("век", "plur"),
    "кв": ("квартал", "sing"),
}
_ROMAN_SHARED_SEPARATOR_PATTERN = re.compile(r"(\s*,\s*|\s+и\s+)")
_ROMAN_CONTEXT_WORD_PATTERN = r"[A-Za-zА-ЯЁа-яё]+\.?"


def _split_context_word_punctuation(word: str) -> tuple[str, str]:
    if word.endswith(".") and normalize_context_token(word) not in _ROMAN_ABBREVIATION_TO_CONTEXT:
        return word[:-1], "."
    return word, ""


def _resolve_roman_context_noun(
    word: str,
    *,
    allow_abbreviations: bool = True,
) -> tuple[str, object, str | None] | None:
    normalized = normalize_context_token(word)
    if not normalized:
        return None
    if allow_abbreviations and normalized in _ROMAN_ABBREVIATION_TO_CONTEXT:
        lemma, abbreviation_number = _ROMAN_ABBREVIATION_TO_CONTEXT[normalized]
        parse = choose_noun_parse(lemma)
        if parse is None:
            return None
        return lemma, parse, abbreviation_number
    parse = choose_noun_parse(word)
    if parse is None or parse.normal_form not in _ROMAN_CONTEXT_LEMMAS:
        return None
    return parse.normal_form, parse, None


def _render_context_noun_word(
    word: str,
    lemma: str,
    noun_parse,
    *,
    case: str,
    abbreviation_number: str | None,
) -> str:
    if abbreviation_number is None:
        return word
    inflected = noun_parse.inflect({case, abbreviation_number})
    return inflected.word if inflected else lemma


def _render_ordinal_for_context_noun(
    number: int,
    noun_parse,
    *,
    case: str | None = None,
) -> str:
    return render_ordinal(
        number,
        case=case or noun_parse_case(noun_parse),
        gender=noun_parse.tag.gender or "masc",
        inanimate="inan" in noun_parse.tag,
    )


def _resolve_explicit_roman_context_form(
    word: str,
    *,
    case_override: str | None = None,
) -> tuple[object, str, str] | None:
    base_word, trailing_punctuation = _split_context_word_punctuation(word)
    resolved = _resolve_roman_context_noun(base_word)
    if resolved is None:
        return None
    lemma, noun_parse, abbreviation_number = resolved
    case = case_override or ("nomn" if abbreviation_number else noun_parse_case(noun_parse))
    rendered_word = _render_context_noun_word(
        base_word,
        lemma,
        noun_parse,
        case=case,
        abbreviation_number=abbreviation_number,
    )
    return noun_parse, case, f"{rendered_word}{trailing_punctuation}"


def _infer_roman_context_case(
    text: str,
    match: re.Match[str],
    noun_word: str,
) -> str:
    left_context = text[max(0, match.start() - 60) : match.start()]
    right_context = text[match.end() : match.end() + 60]
    left_tokens = simple_tokenize(left_context)
    tokens = left_tokens + ["1", noun_word] + simple_tokenize(right_context)
    return get_numeral_case(tokens, len(left_tokens))


def _resolve_full_noun_case_or_none(
    word: str,
    lemma: str,
) -> str | None:
    cases = {
        ("loct" if "loc2" in candidate.tag else (candidate.tag.case or "nomn"))
        for candidate in get_morph().parse(word.lower())
        if "NOUN" in candidate.tag and candidate.normal_form == lemma
    }
    if len(cases) == 1:
        return next(iter(cases))
    return None


def _resolve_contextual_ordinal_case(
    text: str,
    match: re.Match[str],
    noun_word: str,
    lemma: str,
    *,
    abbreviation_number: str | None,
) -> str:
    if abbreviation_number and lemma == "век":
        return _infer_abbreviated_century_case(text, match)
    if abbreviation_number and lemma == "квартал":
        return _infer_abbreviated_quarter_case(text, match)
    if abbreviation_number:
        return _infer_roman_context_case(text, match, lemma)
    explicit_case = _resolve_full_noun_case_or_none(noun_word, lemma)
    if explicit_case is not None:
        return explicit_case
    return _infer_roman_context_case(text, match, noun_word)


def _infer_abbreviated_century_case(
    text: str,
    match: re.Match[str],
) -> str:
    case = _infer_roman_context_case(text, match, "век")
    if case != "accs":
        return case
    left_context = text[max(0, match.start() - 40) : match.start()]
    left_tokens = simple_tokenize(left_context)
    left_word = next(
        (
            normalize_context_token(token)
            for token in reversed(left_tokens)
            if any(char.isalpha() for char in token)
        ),
        "",
    )
    if left_word in {"в", "во", "на", "о", "об", "обо", "при"}:
        return "loct"
    return case


def _infer_abbreviated_quarter_case(
    text: str,
    match: re.Match[str],
) -> str:
    case = _infer_roman_context_case(text, match, "квартал")
    if case != "accs":
        return case
    left_context = text[max(0, match.start() - 40) : match.start()]
    left_tokens = simple_tokenize(left_context)
    left_word = next(
        (
            normalize_context_token(token)
            for token in reversed(left_tokens)
            if any(char.isalpha() for char in token)
        ),
        "",
    )
    if left_word in {"в", "во", "о", "об", "обо", "при"}:
        return "loct"
    return case


def _render_shared_ordinal_in_context(
    text: str,
    match: re.Match[str],
    number: int,
    noun_word: str,
) -> str | None:
    resolved = _resolve_roman_context_noun(noun_word, allow_abbreviations=False)
    if resolved is None:
        return None
    lemma, noun_parse, abbreviation_number = resolved
    case = _resolve_contextual_ordinal_case(
        text,
        match,
        noun_word,
        lemma,
        abbreviation_number=abbreviation_number,
    )
    return _render_ordinal_for_context_noun(number, noun_parse, case=case)


def _render_single_roman_with_context(
    text: str,
    match: re.Match[str],
    number: int,
    noun_word: str,
) -> str | None:
    base_word, trailing_punctuation = _split_context_word_punctuation(noun_word)
    resolved = _resolve_roman_context_noun(base_word)
    if resolved is None:
        return None
    lemma, noun_parse, abbreviation_number = resolved
    case = (
        noun_parse_case(noun_parse)
        if not abbreviation_number
        else _resolve_contextual_ordinal_case(
            text,
            match,
            base_word,
            lemma,
            abbreviation_number=abbreviation_number,
        )
    )
    rendered_noun = _render_context_noun_word(
        base_word,
        lemma,
        noun_parse,
        case=case,
        abbreviation_number=abbreviation_number,
    )
    rendered_ordinal = _render_ordinal_for_context_noun(
        number,
        noun_parse,
        case=case,
    )
    return f"{rendered_ordinal} {rendered_noun}{trailing_punctuation}"


def _render_roman_series_with_context(
    text: str,
    match: re.Match[str],
    series: str,
    noun_word: str,
    *,
    noun_first: bool = False,
) -> str | None:
    base_word, trailing_punctuation = _split_context_word_punctuation(noun_word)
    resolved = _resolve_roman_context_noun(base_word)
    if resolved is None:
        return None
    lemma, noun_parse, abbreviation_number = resolved
    case = _resolve_contextual_ordinal_case(
        text,
        match,
        base_word,
        lemma,
        abbreviation_number=abbreviation_number,
    )
    parts = _ROMAN_SHARED_SEPARATOR_PATTERN.split(series)
    rendered_parts: list[str] = []
    for part in parts:
        if not part:
            continue
        if _ROMAN_SHARED_SEPARATOR_PATTERN.fullmatch(part):
            rendered_parts.append(part)
            continue
        try:
            number = roman.fromRoman(part.upper())
        except roman.InvalidRomanNumeralError:
            return None
        rendered_parts.append(
            _render_ordinal_for_context_noun(number, noun_parse, case=case)
        )
    rendered_series = "".join(rendered_parts)
    rendered_noun = _render_context_noun_word(
        base_word,
        lemma,
        noun_parse,
        case=case,
        abbreviation_number=abbreviation_number,
    )
    if noun_first:
        if noun_word[:1].isupper() and rendered_series:
            rendered_series = rendered_series[:1].upper() + rendered_series[1:]
        return f"{rendered_series} {rendered_noun.lower()}{trailing_punctuation}"
    return f"{rendered_series} {rendered_noun}{trailing_punctuation}"


def _render_roman_hyphen_range_with_context(
    text: str,
    match: re.Match[str],
    left_number: int,
    right_number: int,
    noun_word: str,
    *,
    noun_first: bool = False,
) -> str | None:
    base_word, trailing_punctuation = _split_context_word_punctuation(noun_word)
    resolved = _resolve_roman_context_noun(base_word)
    if resolved is None:
        return None
    lemma, noun_parse, abbreviation_number = resolved
    case = _resolve_contextual_ordinal_case(
        text,
        match,
        base_word,
        lemma,
        abbreviation_number=abbreviation_number,
    )
    rendered_noun = _render_context_noun_word(
        base_word,
        lemma,
        noun_parse,
        case=case,
        abbreviation_number=abbreviation_number,
    )
    left_ordinal = _render_ordinal_for_context_noun(left_number, noun_parse, case=case)
    right_ordinal = _render_ordinal_for_context_noun(right_number, noun_parse, case=case)
    rendered_range = f"{left_ordinal} — {right_ordinal}"
    if noun_first:
        if noun_word[:1].isupper() and rendered_range:
            rendered_range = rendered_range[:1].upper() + rendered_range[1:]
        return f"{rendered_range} {base_word.lower()}{trailing_punctuation}"
    return f"{rendered_range} {rendered_noun}{trailing_punctuation}"


def convert_shared_roman_words(text: str) -> str:
    pattern = re.compile(
        rf"\b(?P<series>[IVXLCDM]+(?:\s*(?:,\s*|\s+и\s+)[IVXLCDM]+)+)\s+(?P<suffix>{_ROMAN_CONTEXT_WORD_PATTERN})(?!\w)"
    )

    def repl(match: re.Match[str]) -> str:
        rendered = _render_roman_series_with_context(
            text,
            match,
            match.group("series"),
            match.group("suffix"),
        )
        return rendered if rendered is not None else match.group(0)

    return pattern.sub(repl, text)


def convert_left_shared_roman_words(text: str) -> str:
    pattern = re.compile(
        rf"\b(?P<suffix>{_ROMAN_CONTEXT_WORD_PATTERN})\s+(?P<series>[IVXLCDM]+(?:\s*(?:,\s*|\s+и\s+)[IVXLCDM]+)+)\b",
        re.IGNORECASE,
    )

    def repl(match: re.Match[str]) -> str:
        rendered = _render_roman_series_with_context(
            text,
            match,
            match.group("series"),
            match.group("suffix"),
            noun_first=True,
        )
        return rendered if rendered is not None else match.group(0)

    return pattern.sub(repl, text)


def convert_roman_words(text: str) -> str:
    pattern = re.compile(
        rf"\b(?P<roman>[IVXLCDM]+)(?P<spacing>\s*)(?P<word>{_ROMAN_CONTEXT_WORD_PATTERN})(?!\w)",
        re.IGNORECASE,
    )

    def repl(match: re.Match[str]) -> str:
        try:
            number = roman.fromRoman(match.group("roman").upper())
        except roman.InvalidRomanNumeralError:
            return match.group(0)
        left_hyphen_context = text[max(0, match.start() - 16) : match.start()]
        if re.search(r"[IVXLCDM]+[-–—]\s*$", left_hyphen_context, re.IGNORECASE):
            return match.group(0)
        if not match.group("spacing") and normalize_context_token(match.group("word")) != "в":
            return match.group(0)
        rendered = _render_single_roman_with_context(
            text,
            match,
            number,
            match.group("word"),
        )
        return rendered if rendered is not None else match.group(0)

    return pattern.sub(repl, text)


def convert_roman_century_ranges(text: str) -> str:
    pattern = re.compile(
        r"\b(?P<prep>с|со|от)\s+(?P<left>[IVXLCDM]+)\s+(?P<mid>до|по)\s+(?P<right>[IVXLCDM]+)\s+(?P<word>век(?:а|у|е|ом|ами|ах|ов)?|в\.|вв\.)(?!\w)",
        re.IGNORECASE,
    )

    def repl(match: re.Match[str]) -> str:
        try:
            left_number = roman.fromRoman(match.group("left").upper())
            right_number = roman.fromRoman(match.group("right").upper())
        except roman.InvalidRomanNumeralError:
            return match.group(0)
        normalized_word = normalize_context_token(match.group("word"))
        if match.group("mid").lower() == "до":
            right_case = "gent"
        elif normalized_word == "вв":
            right_case = "nomn"
        else:
            right_case = None
        resolved = _resolve_explicit_roman_context_form(
            match.group("word"),
            case_override=right_case,
        )
        if resolved is None:
            return match.group(0)
        noun_parse, resolved_right_case, rendered_word = resolved
        return (
            f"{match.group('prep')} {_render_ordinal_for_context_noun(left_number, noun_parse, case='gent')} "
            f"{match.group('mid')} {_render_ordinal_for_context_noun(right_number, noun_parse, case=resolved_right_case)} {rendered_word}"
        )

    return pattern.sub(repl, text)


def convert_hyphenated_roman_ranges(text: str) -> str:
    pattern = re.compile(
        rf"\b(?P<left>[IVXLCDM]+)\s*[-–—]\s*(?P<right>[IVXLCDM]+)\s+(?P<word>{_ROMAN_CONTEXT_WORD_PATTERN})(?!\w)"
    )

    def repl(match: re.Match[str]) -> str:
        try:
            left_number = roman.fromRoman(match.group("left").upper())
            right_number = roman.fromRoman(match.group("right").upper())
        except roman.InvalidRomanNumeralError:
            return match.group(0)
        rendered = _render_roman_hyphen_range_with_context(
            text,
            match,
            left_number,
            right_number,
            match.group("word"),
        )
        return rendered if rendered is not None else match.group(0)

    return pattern.sub(repl, text)


def convert_left_shared_hyphenated_roman_ranges(text: str) -> str:
    pattern = re.compile(
        rf"\b(?P<word>{_ROMAN_CONTEXT_WORD_PATTERN})\s+(?P<left>[IVXLCDM]+)\s*[-–—]\s*(?P<right>[IVXLCDM]+)\b",
        re.IGNORECASE,
    )

    def repl(match: re.Match[str]) -> str:
        try:
            left_number = roman.fromRoman(match.group("left").upper())
            right_number = roman.fromRoman(match.group("right").upper())
        except roman.InvalidRomanNumeralError:
            return match.group(0)
        rendered = _render_roman_hyphen_range_with_context(
            text,
            match,
            left_number,
            right_number,
            match.group("word"),
            noun_first=True,
        )
        return rendered if rendered is not None else match.group(0)

    return pattern.sub(repl, text)


def convert_roman_names(text: str) -> str:
    pattern = r"\b(?P<name>[А-ЯЁ][а-яё]+)\s+(?P<roman>[IVXLCDM]+)\b"

    def pick_name_parse(match: re.Match[str]):
        name = match.group("name")
        parses = [
            candidate
            for candidate in get_morph().parse(name)
            if "NOUN" in candidate.tag
            and "anim" in candidate.tag
            and any(
                marker in candidate.tag for marker in ("Name", "Surn", "Patr")
            )
        ]
        if not parses:
            return None

        left_context = text[max(0, match.start() - 40) : match.start()]
        left_tokens = simple_tokenize(left_context)
        left_word = next(
            (
                normalize_context_token(token)
                for token in reversed(left_tokens)
                if any(char.isalpha() for char in token)
            ),
            "",
        )
        target_case = _REGNAL_CASE_CONTEXT.get(left_word)
        if target_case is not None:
            for candidate in parses:
                candidate_case = "loct" if "loc2" in candidate.tag else candidate.tag.case
                if candidate_case == target_case:
                    return candidate
        return parses[0]

    def repl(match: re.Match[str]) -> str:
        try:
            number = roman.fromRoman(match.group("roman"))
        except roman.InvalidRomanNumeralError:
            return match.group(0)
        parse = pick_name_parse(match)
        if parse is None:
            return match.group(0)
        case = noun_parse_case(parse)
        gender = parse.tag.gender or "masc"
        return f"{match.group('name')} {render_ordinal(number, case=case, gender=gender)}"

    return re.sub(pattern, repl, text)


def convert_heading_roman_numerals(text: str) -> str:
    pattern = re.compile(
        rf"\b(?P<word>{_ROMAN_CONTEXT_WORD_PATTERN})\s+(?P<roman>[IVXLCDM]+)\b",
        re.IGNORECASE,
    )

    def repl(match: re.Match[str]) -> str:
        if match.group("roman") != match.group("roman").upper():
            return match.group(0)
        resolved = _resolve_roman_context_noun(
            match.group("word"),
            allow_abbreviations=False,
        )
        if resolved is None or resolved[0] not in _ROMAN_HEADING_LEMMAS:
            return match.group(0)
        try:
            ordinal = render_ordinal_from_noun_word(
                roman.fromRoman(match.group("roman").upper()),
                match.group("word"),
            )
            if ordinal is None:
                return f"{match.group('word')} {roman.fromRoman(match.group('roman').upper())}"
            return f"{match.group('word')} {ordinal}"
        except roman.InvalidRomanNumeralError:
            return match.group(0)

    return pattern.sub(repl, text)


def convert_other_roman_numerals(text: str) -> str:
    pattern = r"\b([IVXLCDM]{2,})\b"

    def repl(match: re.Match[str]) -> str:
        token = match.group(1).upper()
        left_hyphen_context = text[max(0, match.start() - 16) : match.start()]
        right_hyphen_context = text[match.end() : match.end() + 16]
        left_hyphen_match = re.search(
            r"([IVXLCDM]+)[-–—]\s*$", left_hyphen_context, re.IGNORECASE
        )
        if left_hyphen_match is not None:
            try:
                roman.fromRoman(left_hyphen_match.group(1).upper())
            except roman.InvalidRomanNumeralError:
                return match.group(0)
        right_hyphen_match = re.match(
            r"\s*[-–—]([IVXLCDM]+)", right_hyphen_context, re.IGNORECASE
        )
        if right_hyphen_match is not None:
            try:
                roman.fromRoman(right_hyphen_match.group(1).upper())
            except roman.InvalidRomanNumeralError:
                return match.group(0)
        if token in _ROMAN_ABBREVIATION_EXCEPTIONS:
            return match.group(0)
        try:
            return str(roman.fromRoman(token))
        except roman.InvalidRomanNumeralError:
            return match.group(0)

    return re.sub(pattern, repl, text)


def normalize_cyrillic_roman(text: str) -> str:
    cyrillic_to_latin = {
        "Х": "X",
        "х": "X",
        "С": "C",
        "с": "C",
        "І": "I",
        "l": "I",
        "і": "I",
        "М": "M",
        "м": "M",
    }
    pattern = re.compile(r"\b([IVXLCDMХхСсІіМм]{2,})\b")

    def repl(match: re.Match[str]) -> str:
        word = match.group(1)
        if word == word.lower():
            return word
        if word != word.upper():
            return word
        if is_safe_numeric_hyphen_unit(word):
            left_context = text[max(0, match.start() - 32) : match.start()]
            if re.search(r"\d+(?:[.,]\d+)?\s*[-–—]?\s*$", left_context):
                return word
        if not any(char in "ХхСсІіМм" for char in word):
            return word
        latin_word = "".join(cyrillic_to_latin.get(char, char) for char in word).upper()
        try:
            roman.fromRoman(latin_word)
            return latin_word
        except roman.InvalidRomanNumeralError:
            return word

    return pattern.sub(repl, text)


def normalize_roman(text: str, options: NormalizeOptions | None = None) -> str:
    active = options or NormalizeOptions()
    if not active.enable_roman_normalization:
        return text
    text = normalize_cyrillic_roman(text)
    text = convert_roman_century_ranges(text)
    text = convert_hyphenated_roman_ranges(text)
    text = convert_shared_roman_words(text)
    text = convert_left_shared_hyphenated_roman_ranges(text)
    text = convert_left_shared_roman_words(text)
    text = convert_roman_words(text)
    text = convert_roman_names(text)
    text = convert_heading_roman_numerals(text)
    text = convert_other_roman_numerals(text)
    return text



from __future__ import annotations

import re

RU_LETTER_NAMES = {
    "А": "аа",
    "Б": "бэ",
    "В": "вэ",
    "Г": "гэ",
    "Д": "дэ",
    "Е": "ее",
    "Ё": "ё",
    "Ж": "жэ",
    "З": "зэ",
    "И": "ии",
    "Й": "й",
    "К": "ка",
    "Л": "эл",
    "М": "эм",
    "Н": "эн",
    "О": "оо",
    "П": "пэ",
    "Р": "эр",
    "С": "эс",
    "Т": "тэ",
    "У": "уу",
    "Ф": "эф",
    "Х": "ха",
    "Ц": "цэ",
    "Ч": "чэ",
    "Ш": "шэ",
    "Щ": "ща",
    "Ъ": "",
    "Ы": "ыы",
    "Ь": "",
    "Э": "ээ",
    "Ю": "юю",
    "Я": "яя",
}

EN_LETTER_NAMES = {
    "A": "эй",
    "B": "би",
    "C": "си",
    "D": "ди",
    "E": "ии",
    "F": "эф",
    "G": "джи",
    "H": "эйч",
    "I": "ай",
    "J": "джей",
    "K": "кей",
    "L": "эл",
    "M": "эм",
    "N": "эн",
    "O": "оу",
    "P": "пи",
    "Q": "кью",
    "R": "ар",
    "S": "эс",
    "T": "ти",
    "U": "юю",
    "V": "ви",
    "W": "дабел ю",
    "X": "экс",
    "Y": "уай",
    "Z": "зи",
}

KNOWN_ABBREVIATIONS = {
    "ЛСД": "лсд",
    "СПИД": "спид",
    "ВИЧ": "вич",
    "ВУЗ": "вуз",
    "СУЗ": "суз",
    "НАТО": "нато",
    "ЮНЕСКО": "юнеско",
    "ОПЕК": "опек",
    "ФИФА": "фифа",
    "УЕФА": "уефа",
    "НАСА": "наса",
    "РОСКОСМОС": "роскосмос",
    "МАГАТЭ": "магатэ",
    "ОМОН": "омон",
    "СОБР": "собр",
    "ЗАГС": "загс",
    "СМИ": "сми",
    "ГОСТ": "гост",
    "МХАТ": "мхат",
    "ЕГЭ": "егэ",
    "ОГЭ": "огэ",
    "ГИБДД": "ги бэ дэ дэ",
    "ТАСС": "тасс",
    "КАМАЗ": "камаз",
    "ВАЗ": "ваз",
    "ГАЗ": "газ",
    "УАЗ": "уаз",
    "ЗИЛ": "зил",
    "КРАЗ": "краз",
    "МАЗ": "маз",
    "БЕЛАЗ": "белаз",
    "ЛУАЗ": "луаз",
    "WI-FI": "вайф+ай",
    "WIFI": "вайф+ай",
    "VIP": "ви ай пи",
    "PS": "пост скр+иптум",
    "SOS": "сос",
    "IT": "ай ти",
    "TV": "ти ви",
    "СССР": "эс эс эс эр",
    "COVID": "ковид",
    "НЕТ": "нет",
    "ДА": "да",
    "МИД": "мид",
    "США": "сэ шэ аа",
    "РФ": "эр эф",
    "ООН": "о он",
    "ВВП": "вэ вэ пэ",
    "ЧС": "че эс",
    "МЧС": "эм че эс",
    "ФСБ": "эф эс бэ",
    "МВД": "эм вэ дэ",
    "ГРУ": "гэ рэ у",
    "ДНК": "дэ эн ка",
    "ЖКХ": "же кэ ха",
    "ТЭС": "тэс",
    "ГЭС": "гэс",
    "АЭС": "аэс",
    "ВЦИОМ": "вци+ом",
    "КПСС": "ка пэ эс эс",
    "ВЛКСМ": "вэ эл ка эс эм",
    "ЦК": "цэ ка",
    "РСФСР": "эр эс эф эс эр",
    "УССР": "у эс эс эр",
    "БССР": "бэ эс эс эр",
    "СНГ": "эс эн гэ",
    "ЕЭС": "е ээ эс",
    "ЕС": "е эс",
    "ВОЗ": "воз",
    "МОК": "мок",
    "НХЛ": "эн ха эл",
    "НБА": "эн бэ аа",
    "МВФ": "эм вэ эф",
    "ВТО": "вэ тэ оо",
    "ДПС": "дэ пэ эс",
    "ППС": "пэ пэ эс",
    "УФСИН": "уфсин",
    "ФСИН": "фсин",
    "МФЦ": "эм эф цэ",
    "ПФР": "пэ фэ эр",
    "ФНС": "эф эн эс",
    "ФАС": "фас",
    "ЦБ": "цэ бэ",
    "РЖД": "эр же дэ",
    "АЗС": "а зэ эс",
    "ТЦ": "тэ цэ",
    "ТРЦ": "тэ эр цэ",
    "БЦ": "бэ цэ",
    "ЖК": "же ка",
    "ТСЖ": "тэ сэ жэ",
    "СНТ": "эс эн тэ",
    "ОАО": "оо аа оо",
    "ЗАО": "зз аа оо",
    "ООО": "оо оо оо",
    "ИП": "и пэ",
    "ГБУ": "гэ бэ у",
    "МБУ": "эм бэ у",
    "ФГБУ": "эф гэ бэ у",
    "НИИ": "ни и",
    "ПТУ": "пэ тэ у",
    "ВПР": "вэ пэ эр",
    "ГТО": "гэ тэ о",
    "КВН": "ка вэ эн",
    "ЧП": "че пэ",
    "ДТП": "дэ тэ пэ",
    "ТВ": "тэ вэ",
    "NB": "эн би",
    "SMS": "эс эм эс",
    "SIM": "сим",
    "PIN": "пин",
    "PUK": "пак",
    "ID": "ай ди",
    "USB": "ю эс би",
    "CD": "си ди",
    "DVD": "ди ви ди",
    "GPS": "джи пи эс",
    "ОРВИ": "оэрв+и",
    "ОРЗ": "оо эр зэ",
    "МРОТ": "мрот",
    "СНИЛС": "снилс",
    "АВТОВАЗ": "автоваз",
}

RU_VOWELS = set("АЕЁИОУЫЭЮЯ")
ABBREV_MIN_LEN = 2
ABBREV_MAX_LEN_RU = 5
ABBREV_MAX_LEN_EN = 5
REAL_WORD_POS = frozenset(
    {
        "NOUN",
        "VERB",
        "ADJF",
        "ADJS",
        "ADVB",
        "INFN",
        "PRTF",
        "PRTS",
        "GRND",
        "PREP",
        "CONJ",
        "PRCL",
        "NPRO",
    }
)

RUSSIAN_NAME_TOKEN = r"[А-ЯЁ][а-яё+]+(?:-[А-ЯЁ]?[а-яё+]+)*"
PERSON_SURNAME_INITIALS_PATTERN = re.compile(
    rf"(?<![A-Za-zА-Яа-яЁё])(?P<surname>{RUSSIAN_NAME_TOKEN})\s+"
    r"(?P<i1>[А-ЯЁ])\.\s*(?P<i2>[А-ЯЁ])\.(?![A-Za-zА-Яа-яЁё])"
)
PERSON_INITIALS_SURNAME_PATTERN = re.compile(
    rf"(?<![A-Za-zА-Яа-яЁё])(?P<i1>[А-ЯЁ])\.\s*(?P<i2>[А-ЯЁ])\.\s+"
    rf"(?P<surname>{RUSSIAN_NAME_TOKEN})(?![A-Za-zА-Яа-яЁё])"
)
PERSON_SURNAME_SINGLE_INITIAL_PATTERN = re.compile(
    rf"(?<![A-Za-zА-Яа-яЁё])(?P<surname>{RUSSIAN_NAME_TOKEN})\s+"
    r"(?P<i1>[А-ЯЁ])\.(?![A-Za-zА-Яа-яЁё])"
)
PERSON_SINGLE_INITIAL_SURNAME_PATTERN = re.compile(
    rf"(?<![A-Za-zА-Яа-яЁё])(?P<i1>[А-ЯЁ])\.\s+"
    rf"(?P<surname>{RUSSIAN_NAME_TOKEN})(?![A-Za-zА-Яа-яЁё])"
)

ABBREV_DOTTED_PATTERN = re.compile(
    r"(?<![A-Za-zА-ЯЁ])(?:[A-ZА-ЯЁ]\.){2,}[A-ZА-ЯЁ]?\.?(?![A-Za-zА-ЯЁ])"
)
ABBREV_SOLID_PATTERN = re.compile(r"(?<![A-Za-zА-ЯЁ])[A-ZА-ЯЁ]{2,}(?![A-Za-zА-ЯЁ])")

UNICODE_FRACTIONS = {
    "¼": "1/4",
    "½": "1/2",
    "¾": "3/4",
    "⅓": "1/3",
    "⅔": "2/3",
    "⅛": "1/8",
    "⅜": "3/8",
    "⅝": "5/8",
    "⅞": "7/8",
}

CLEANUP_REPLACEMENTS = [
    ("--", "—"),
    (",,", ","),
    (".\n", ". \n"),
    ("©", "."),
    ("®", ""),
    ("™", ""),
    ("−", "-"),
    ("–", "—"),
    ("\u2010", "-"),
    ("\u2011", "-"),
    ("\u2212", "-"),
    ("-", "-"),
    ("\u00a0", " "),
    ("\u202f", " "),
    ("\u2003", " "),
    ("\u2002", " "),
    ("\u2007", " "),
    ("\u2008", " "),
    ("\u2009", " "),
    ("\t", " "),
]



from __future__ import annotations

import re

POST_NUMERAL_ABBREVIATION_PATTERNS = [
    (re.compile(r"(?<!\w)тыс\.(?!\w)", re.IGNORECASE), "тысяч"),
    (re.compile(r"(?<!\w)млн\.(?!\w)", re.IGNORECASE), "миллионов"),
    (re.compile(r"(?<!\w)млрд\.(?!\w)", re.IGNORECASE), "миллиардов"),
    (re.compile(r"(?<!\w)трлн\.(?!\w)", re.IGNORECASE), "триллионов"),
    (re.compile(r"(?<!\w)руб\.(?!\w)", re.IGNORECASE), "рублей"),
    (re.compile(r"(?<!\w)долл\.(?!\w)", re.IGNORECASE), "долларов"),
    (re.compile(r"(?<!\w)дол\.(?!\w)", re.IGNORECASE), "долларов"),
    (re.compile(r"(?<!\w)и\s+др\.(?!\w)", re.IGNORECASE), "и другие"),
]

GREEK_LETTERS = {
    "α": "альфа",
    "β": "бета",
    "γ": "гамма",
    "δ": "дельта",
    "ε": "эпсилон",
    "ϵ": "эпсилон",
    "ζ": "дзета",
    "η": "эта",
    "θ": "тета",
    "ϑ": "тета",
    "ι": "йота",
    "κ": "каппа",
    "ϰ": "каппа",
    "λ": "лямбда",
    "μ": "мю",
    "µ": "мю",
    "ν": "ню",
    "ξ": "кси",
    "ο": "омикрон",
    "π": "пи",
    "ϖ": "пи",
    "ρ": "ро",
    "ϱ": "ро",
    "σ": "сигма",
    "ς": "сигма",
    "τ": "тау",
    "φ": "фи",
    "ϕ": "фи",
    "χ": "хи",
    "ψ": "пси",
    "ω": "омега",
    "Δ": "дельта",
    "Σ": "сигма",
    "Ω": "омега",
    "Π": "пи",
    "Φ": "фи",
    "Γ": "гамма",
    "Λ": "лямбда",
    "Θ": "тета",
}

MATH_SYMBOLS = {
    "×": " умножить на ",
    "÷": " разделить на ",
    "±": " плюс-минус ",
    "∓": " минус-плюс ",
    "≈": " приблизительно равно ",
    "≠": " не равно ",
    "≤": " меньше или равно ",
    "≥": " больше или равно ",
    "∞": " бесконечность ",
    "√": " корень из ",
    "∑": " сумма ",
    "∏": " произведение ",
    "∫": " интеграл ",
    "∂": " дель ",
    "∈": " принадлежит ",
    "∉": " не принадлежит ",
    "∪": " объединение ",
    "∩": " пересечение ",
    "⇒": " следует ",
    "№": " номер ",
}

CURRENCY_STANDALONE = {
    "$": "доллар",
    "€": "евро",
    "£": "фунт",
    "¥": "йена",
    "₽": "рубль",
    "₴": "гривна",
    "₸": "тенге",
    "₺": "лира",
    "₹": "рупия",
    "¢": "цент",
}

PREP_CASE = {
    "в": "accs",
    "о": "loct",
    "об": "loct",
    "на": "accs",
    "за": "accs",
    "из": "gent",
    "с": "gent",
    "к": "datv",
    "по": "datv",
    "для": "gent",
    "без": "gent",
    "при": "loct",
    "между": "ablt",
    "до": "gent",
    "от": "gent",
    "у": "gent",
    "во": "accs",
    "со": "gent",
    "ко": "datv",
    "обо": "loct",
    "около": "gent",
    "порядка": "gent",
    "не более": "gent",
    "не менее": "gent",
    "не больше": "gent",
    "не меньше": "gent",
    "более": "gent",
    "больше": "gent",
    "менее": "gent",
    "меньше": "gent",
    "свыше": "gent",
    "после": "gent",
    "против": "gent",
    "среди": "gent",
    "вместо": "gent",
    "возле": "gent",
    "вокруг": "gent",
    "внутри": "gent",
    "вне": "gent",
    "ради": "gent",
    "в течение": "gent",
    "течение": "gent",
    "в продолжение": "gent",
    "из-за": "gent",
    "из-под": "gent",
    "вблизи": "gent",
    "вроде": "gent",
    "наподобие": "gent",
    "касательно": "gent",
    "относительно": "gent",
    "насчет": "gent",
    "по поводу": "gent",
    "по причине": "gent",
    "ввиду": "gent",
    "в силу": "gent",
    "в целях": "gent",
    "в рамках": "gent",
    "в ходе": "gent",
    "в отсутствие": "gent",
    "во избежание": "gent",
    "по истечении": "gent",
    "исходя из": "gent",
    "независимо от": "gent",
    "в отличие от": "gent",
    "в зависимости от": "gent",
    "вследствие": "gent",
    "через": "accs",
    "вдоль": "gent",
    "благодаря": "datv",
    "вопреки": "datv",
    "согласно": "datv",
    "навстречу": "datv",
    "подобно": "datv",
    "сообразно": "datv",
    "применительно к": "datv",
    "по направлению к": "datv",
    "по мере": "gent",
    "над": "ablt",
    "надо": "ablt",
    "перед": "ablt",
    "передо": "ablt",
    "пред": "ablt",
    "меж": "ablt",
    "в связи с": "ablt",
    "рядом с": "ablt",
    "вместе с": "ablt",
    "наряду с": "ablt",
    "старше": "gent",
    "моложе": "gent",
    "младше": "gent",
    "выше": "gent",
    "ниже": "gent",
    "дольше": "gent",
    "короче": "gent",
}

VERB_CASE = {
    "обнаружить": "accs",
    "выявить": "accs",
    "показать": "accs",
    "считать": "accs",
    "оценить": "accs",
    "иметь": "accs",
    "избегать": "gent",
    "опасаться": "gent",
    "достичь": "gent",
    "достигать": "gent",
    "лишиться": "gent",
    "верить": "datv",
    "доверять": "datv",
    "видеть": "accs",
    "наблюдать": "accs",
    "замечать": "accs",
    "осознавать": "accs",
    "понимать": "accs",
    "изучать": "accs",
    "анализировать": "accs",
    "рассматривать": "accs",
    "определять": "accs",
    "устанавливать": "accs",
    "формировать": "accs",
    "использовать": "accs",
    "применять": "accs",
    "реализовывать": "accs",
    "разрабатывать": "accs",
    "создавать": "accs",
    "включать": "accs",
    "получать": "accs",
    "не иметь": "gent",
    "избежать": "gent",
    "достигнуть": "gent",
    "требовать": "gent",
    "ожидать": "gent",
    "бояться": "gent",
    "страшиться": "gent",
    "лишать": "gent",
    "касаться": "gent",
    "относиться": "gent",
    "состоять": "gent",
    "дожидаться": "gent",
    "избавиться": "gent",
    "помогать": "datv",
    "способствовать": "datv",
    "препятствовать": "datv",
    "следовать": "datv",
    "соответствовать": "datv",
    "противоречить": "datv",
    "удивляться": "datv",
    "радоваться": "datv",
    "завидовать": "datv",
    "угрожать": "datv",
    "рекомендовать": "datv",
    "предлагать": "datv",
    "являться": "ablt",
    "считаться": "ablt",
    "обладать": "ablt",
    "характеризоваться": "ablt",
    "отличаться": "ablt",
    "определяться": "ablt",
    "пользоваться": "ablt",
    "управлять": "ablt",
    "владеть": "ablt",
    "говорить": "loct",
    "сообщать": "loct",
    "рассказывать": "loct",
    "упоминать": "loct",
    "писать": "loct",
    "думать": "loct",
    "размышлять": "loct",
    "основываться": "loct",
    "заключаться": "loct",
}

UNITS_DATA = {
    "₽": ("рубль", "masc", "money"),
    "р": ("рубль", "masc", "money"),
    "руб": ("рубль", "masc", "money"),
    "руб.": ("рубль", "masc", "money"),
    "рубль": ("рубль", "masc", "money"),
    "$": ("доллар", "masc", "money"),
    "¢": ("цент", "masc", "money"),
    "доллар": ("доллар", "masc", "money"),
    "долл.": ("доллар", "masc", "money"),
    "долл": ("доллар", "masc", "money"),
    "usd": ("доллар", "masc", "money"),
    "€": ("евро", "masc", "money"),
    "eur": ("евро", "masc", "money"),
    "евро": ("евро", "masc", "money"),
    "£": ("фунт", "masc", "money"),
    "фунт": ("фунт", "masc", "money"),
    "¥": ("иена", "femn", "money"),
    "иена": ("иена", "femn", "money"),
    "¥н": ("юань", "masc", "money"),
    "юань": ("юань", "masc", "money"),
    "₪": ("шекель", "masc", "money"),
    "шекель": ("шекель", "masc", "money"),
    "₴": ("гривна", "femn", "money"),
    "uah": ("гривна", "femn", "money"),
    "₸": ("тенге", "masc", "money"),
    "kzt": ("тенге", "masc", "money"),
    "₺": ("лира", "femn", "money"),
    "try": ("лира", "femn", "money"),
    "₹": ("рупия", "femn", "money"),
    "inr": ("рупия", "femn", "money"),
    "₩": ("вон", "masc", "money"),
    "krw": ("вон", "masc", "money"),
    "₫": ("донг", "masc", "money"),
    "vnd": ("донг", "masc", "money"),
    "₱": ("песо", "neut", "money"),
    "₦": ("найра", "femn", "money"),
    "ngn": ("найра", "femn", "money"),
    "копейка": ("копейка", "femn", "money"),
    "коп": ("копейка", "femn", "money"),
    "коп.": ("копейка", "femn", "money"),
    "цент": ("цент", "masc", "money"),
    "%": ("процент", "masc", "percent"),
    "процент": ("процент", "masc", "percent"),
    "‰": ("промилле", "masc", "percent"),
    "‱": ("базисный пункт", "masc", "percent"),
    "п.п.": ("процентный пункт", "masc", "percent"),
    "б.п.": ("базисный пункт", "masc", "percent"),
    "bps": ("базисный пункт", "masc", "percent"),
    "°c": ("градус", "masc", "measure", "Цельсия"),
    "°k": ("градус", "masc", "measure", "Кельвина"),
    "°f": ("градус", "masc", "measure", "Фаренгейта"),
    "°С": ("градус", "masc", "measure", "Цельсия"), # russian C
    "°К": ("градус", "masc", "measure", "Кельвина"), # russian K
    "°": ("градус", "masc", "measure"),
    "градус": ("градус", "masc", "measure"),
    # "k": ("кельвин", "masc", "measure"),
    # "к": ("кельвин", "masc", "measure"),
    "кельвин": ("кельвин", "masc", "measure"),
    "v": ("вольт", "masc", "measure"),
    "в.": ("век", "masc", "measure"),
    "вольт": ("вольт", "masc", "measure"),
    "кв": ("киловольт", "masc", "measure"),
    "mv": ("милливольт", "masc", "measure"),
    "kv": ("киловольт", "masc", "measure"),
    "ma": ("миллиампер", "masc", "measure"),
    "ka": ("килоампер", "masc", "measure"),
    "ua": ("микроампер", "masc", "measure"),
    "µa": ("микроампер", "masc", "measure"),
    "μa": ("микроампер", "masc", "measure"),
    "вт": ("ватт", "masc", "measure"),
    "ватт": ("ватт", "masc", "measure"),
    "w": ("ватт", "masc", "measure"),
    "kw": ("киловатт", "masc", "measure"),
    "mw": ("мегаватт", "masc", "measure"),
    "gw": ("гигаватт", "masc", "measure"),
    "мкА": ("микроампер", "masc", "measure"),
    "мА": ("миллиампер", "masc", "measure"),
    "кА": ("килоампер", "masc", "measure"),
    "ампер": ("ампер", "masc", "measure"),
    "ω": ("ом", "masc", "measure"),
    "ом": ("ом", "masc", "measure"),
    "н": ("ньютон", "masc", "measure"),
    "n": ("ньютон", "masc", "measure"),
    "ньютон": ("ньютон", "masc", "measure"),
    "кн": ("килоньютон", "masc", "measure"),
    "kn": ("килоньютон", "masc", "measure"),
    "мн": ("меганьютон", "masc", "measure"),
    "mn": ("меганьютон", "masc", "measure"),
    "ф": ("фарад", "masc", "measure"),
    "f": ("фарад", "masc", "measure"),
    "фарад": ("фарад", "masc", "measure"),
    "мф": ("миллифарад", "masc", "measure"),
    "mf": ("миллифарад", "masc", "measure"),
    "мкф": ("микрофарад", "masc", "measure"),
    "µf": ("микрофарад", "masc", "measure"),
    "μf": ("микрофарад", "masc", "measure"),
    "uf": ("микрофарад", "masc", "measure"),
    "нф": ("нанофарад", "masc", "measure"),
    "nf": ("нанофарад", "masc", "measure"),
    "пф": ("пикофарад", "masc", "measure"),
    "pf": ("пикофарад", "masc", "measure"),
    "гн": ("генри", "masc", "measure"),
    "h": ("генри", "masc", "measure"),
    "генри": ("генри", "masc", "measure"),
    "кл": ("кулон", "masc", "measure"),
    "c": ("кулон", "masc", "measure"),
    "кулон": ("кулон", "masc", "measure"),
    "тл": ("тесла", "femn", "measure"),
    "тесла": ("тесла", "femn", "measure"),
    "вб": ("вебер", "masc", "measure"),
    "wb": ("вебер", "masc", "measure"),
    "вебер": ("вебер", "masc", "measure"),
    "лм": ("люмен", "masc", "measure"),
    "lm": ("люмен", "masc", "measure"),
    "люмен": ("люмен", "masc", "measure"),
    "лк": ("люкс", "masc", "measure"),
    "lx": ("люкс", "masc", "measure"),
    "люкс": ("люкс", "masc", "measure"),
    "бк": ("беккерель", "masc", "measure"),
    "bq": ("беккерель", "masc", "measure"),
    "беккерель": ("беккерель", "masc", "measure"),
    "gy": ("грей", "masc", "measure"),
    "грей": ("грей", "masc", "measure"),
    "зв": ("зиверт", "masc", "measure"),
    "sv": ("зиверт", "masc", "measure"),
    "зиверт": ("зиверт", "masc", "measure"),
    "дж": ("джоуль", "masc", "measure"),
    "джоуль": ("джоуль", "masc", "measure"),
    "кдж": ("килоджоуль", "masc", "measure"),
    "па": ("паскаль", "masc", "measure"),
    "паскаль": ("паскаль", "masc", "measure"),
    "мпа": ("мегапаскаль", "masc", "measure"),
    "кпа": ("килопаскаль", "masc", "measure"),
    "бар": ("бар", "masc", "measure"),
    "квт": ("киловатт", "masc", "measure"),
    "киловатт": ("киловатт", "masc", "measure"),
    "квтч": ("киловатт-час", "masc", "measure"),
    "квт.ч": ("киловатт-час", "masc", "measure"),
    "квт-ч": ("киловатт-час", "masc", "measure"),
    "мвт": ("мегаватт", "masc", "measure"),
    "гвт": ("гигаватт", "masc", "measure"),
    "гц": ("герц", "masc", "measure"),
    "герц": ("герц", "masc", "measure"),
    "кгц": ("килогерц", "masc", "measure"),
    "мгц": ("мегагерц", "masc", "measure"),
    "ггц": ("гигагерц", "masc", "measure"),
    "байт": ("байт", "masc", "measure"),
    "кб": ("килобайт", "masc", "measure"),
    "кбайт": ("килобайт", "masc", "measure"),
    "мб": ("мегабайт", "masc", "measure"),
    "мбайт": ("мегабайт", "masc", "measure"),
    "гб": ("гигабайт", "masc", "measure"),
    "гбайт": ("гигабайт", "masc", "measure"),
    "тб": ("терабайт", "masc", "measure"),
    "тбайт": ("терабайт", "masc", "measure"),
    "бит": ("бит", "masc", "measure"),
    "кбит": ("килобит", "masc", "measure"),
    "мбит": ("мегабит", "masc", "measure"),
    "гбит": ("гигабит", "masc", "measure"),
    "тбит": ("терабит", "masc", "measure"),
    "kb": ("килобайт", "masc", "measure"),
    "mb": ("мегабайт", "masc", "measure"),
    "gb": ("гигабайт", "masc", "measure"),
    "tb": ("терабайт", "masc", "measure"),
    "kbps": ("килобит", "masc", "measure", "в секунду"),
    "mbps": ("мегабит", "masc", "measure", "в секунду"),
    "gbps": ("гигабит", "masc", "measure", "в секунду"),
    "бит/с": ("бит", "masc", "measure", "в секунду"),
    "кбит/с": ("килобит", "masc", "measure", "в секунду"),
    "мбит/с": ("мегабит", "masc", "measure", "в секунду"),
    "гбит/с": ("гигабит", "masc", "measure", "в секунду"),
    "байт/с": ("байт", "masc", "measure", "в секунду"),
    "кб/с": ("килобайт", "masc", "measure", "в секунду"),
    "мб/с": ("мегабайт", "masc", "measure", "в секунду"),
    "гб/с": ("гигабайт", "masc", "measure", "в секунду"),
    "kb/s": ("килобайт", "masc", "measure", "в секунду"),
    "mb/s": ("мегабайт", "masc", "measure", "в секунду"),
    "gb/s": ("гигабайт", "masc", "measure", "в секунду"),
    "fps": ("кадр", "masc", "measure", "в секунду"),
    "га": ("гектар", "masc", "measure"),
    "гектар": ("гектар", "masc", "measure"),
    "кал": ("калория", "femn", "measure"),
    "ккал": ("килокалория", "femn", "measure"),
    "л.с.": ("лошадиная сила", "femn", "measure"),
    "л. с.": ("лошадиная сила", "femn", "measure"),
    "кг": ("килограмм", "masc", "measure"),
    "килограмм": ("килограмм", "masc", "measure"),
    "kg": ("килограмм", "masc", "measure"),
    "г": ("грамм", "masc", "measure"),
    "грамм": ("грамм", "masc", "measure"),
    "гр": ("грамм", "masc", "measure"),
    "т": ("тонна", "femn", "measure"),
    "тонна": ("тонна", "femn", "measure"),
    "мг": ("миллиграмм", "masc", "measure"),
    "мкг": ("микрограмм", "masc", "measure"),
    "нг": ("нанограмм", "masc", "measure"),
    "oz": ("унция", "femn", "measure"),
    "унция": ("унция", "femn", "measure"),
    "lb": ("фунт", "masc", "measure"),
    "м": ("метр", "masc", "measure"),
    "нм.": ("нанометр", "masc", "measure"),
    "нм": ("нанометр", "masc", "measure"),
    "метр": ("метр", "masc", "measure"),
    "см": ("сантиметр", "masc", "measure"),
    "cm": ("сантиметр", "masc", "measure"),
    "см²": ("квадратный сантиметр", "masc", "measure"),
    "см³": ("кубический сантиметр", "masc", "measure"),
    "см2": ("квадратный сантиметр", "masc", "measure"),
    "см3": ("кубический сантиметр", "masc", "measure"),
    "см^2": ("квадратный сантиметр", "masc", "measure"),
    "см^3": ("кубический сантиметр", "masc", "measure"),
    "мм": ("миллиметр", "masc", "measure"),
    "mm": ("миллиметр", "masc", "measure"),
    "мм²": ("квадратный миллиметр", "masc", "measure"),
    "мм³": ("кубический миллиметр", "masc", "measure"),
    "мм2": ("квадратный миллиметр", "masc", "measure"),
    "мм3": ("кубический миллиметр", "masc", "measure"),
    "мм^2": ("квадратный миллиметр", "masc", "measure"),
    "мм^3": ("кубический миллиметр", "masc", "measure"),
    "м²": ("квадратный метр", "masc", "measure"),
    "м³": ("кубический метр", "masc", "measure"),
    "м2": ("квадратный метр", "masc", "measure"),
    "м3": ("кубический метр", "masc", "measure"),
    "м^2": ("квадратный метр", "masc", "measure"),
    "м^3": ("кубический метр", "masc", "measure"),
    "кв.м": ("квадратный метр", "masc", "measure"),
    "квм": ("квадратный метр", "masc", "measure"),
    "куб.м": ("кубический метр", "masc", "measure"),
    "кубм": ("кубический метр", "masc", "measure"),
    "кмч": ("километр", "masc", "measure", "в час"),
    "км": ("километр", "masc", "measure"),
    "километр": ("километр", "masc", "measure"),
    "km": ("километр", "masc", "measure"),
    "км²": ("квадратный километр", "masc", "measure"),
    "км2": ("квадратный километр", "masc", "measure"),
    "км^2": ("квадратный километр", "masc", "measure"),
    "inch": ("дюйм", "masc", "measure"),
    "inches": ("дюйм", "masc", "measure"),
    "in³": ("кубический дюйм", "masc", "measure"),
    "л": ("литр", "masc", "measure"),
    "литр": ("литр", "masc", "measure"),
    "мл": ("миллилитр", "masc", "measure"),
    "ml": ("миллилитр", "masc", "measure"),
    "миллилитр": ("миллилитр", "masc", "measure"),
    "пинта": ("пинта", "femn", "measure"),
    "pint": ("пинта", "femn", "measure"),
    "м/с": ("метр", "masc", "measure", "в секунду"),
    "км/ч": ("километр", "masc", "measure", "в час"),
    "об/мин": ("оборот", "masc", "measure", "в минуту"),
    "rpm": ("оборот", "masc", "measure", "в минуту"),
    "ч": ("час", "masc", "measure"),
    "ч.": ("час", "masc", "measure"),
    "час": ("час", "masc", "measure"),
    "мин": ("минута", "femn", "measure"),
    "мин.": ("минута", "femn", "measure"),
    "минута": ("минута", "femn", "measure"),
    "сек": ("секунда", "femn", "measure"),
    "сек.": ("секунда", "femn", "measure"),
    "секунд": ("секунда", "femn", "measure"),
    "секунда": ("секунда", "femn", "measure"),
    "с": ("секунда", "femn", "measure"),
    "с.": ("секунда", "femn", "measure"),
    "мс": ("миллисекунда", "femn", "measure"),
    "мкс": ("микросекунда", "femn", "measure"),
    "нс": ("наносекунда", "femn", "measure"),
    "сут": ("сутки", "plur", "measure"),
    "сут.": ("сутки", "plur", "measure"),
    "дн": ("день", "masc", "measure"),
    "дн.": ("день", "masc", "measure"),
    "нед": ("неделя", "femn", "measure"),
    "нед.": ("неделя", "femn", "measure"),
    "мес": ("месяц", "masc", "measure"),
    "мес.": ("месяц", "masc", "measure"),
    "фут": ("фут", "masc", "measure"),
    "фута": ("фут", "masc", "measure"),
    "футов": ("фут", "masc", "measure"),
    "шт": ("штука", "femn", "measure"),
    "шт.": ("штука", "femn", "measure"),
    "штука": ("штука", "femn", "measure"),
    "экземпляр": ("экземпляр", "masc", "measure"),
    "человек": ("человек", "masc", "measure"),
    "единица": ("единица", "femn", "measure"),
    "чел": ("человек", "masc", "measure"),
    "чел.": ("человек", "masc", "measure"),
    "тыс": ("тысяча", "femn", "measure"),
    "тыс.": ("тысяча", "femn", "measure"),
    "тысяча": ("тысяча", "femn", "measure"),
    "тысячи": ("тысяча", "femn", "measure"),
    "тысяч": ("тысяча", "femn", "measure"),
    "млн": ("миллион", "masc", "measure"),
    "млн.": ("миллион", "masc", "measure"),
    "миллион": ("миллион", "masc", "measure"),
    "миллиона": ("миллион", "masc", "measure"),
    "миллионов": ("миллион", "masc", "measure"),
    "млрд": ("миллиард", "masc", "measure"),
    "млрд.": ("миллиард", "masc", "measure"),
    "миллиард": ("миллиард", "masc", "measure"),
    "миллиарда": ("миллиард", "masc", "measure"),
    "миллиардов": ("миллиард", "masc", "measure"),
    "трлн": ("триллион", "masc", "measure"),
    "трлн.": ("триллион", "masc", "measure"),
    "триллион": ("триллион", "masc", "measure"),
    "триллиона": ("триллион", "masc", "measure"),
    "триллионов": ("триллион", "masc", "measure"),
    "ар": ("ар", "masc", "measure"),
    "дм": ("дециметр", "masc", "measure"),
    "ft": ("фут", "masc", "measure"),
    "yd": ("ярд", "masc", "measure"),
    "mi": ("миля", "femn", "measure"),
    "аршин": ("аршин", "masc", "measure"),
    "арш": ("аршин", "masc", "measure"),
    "mph": ("миля", "femn", "measure", "в час"),
    "обмин": ("оборот", "masc", "measure", "в минуту"),
    "уз": ("узел", "masc", "measure"),
    "узел": ("узел", "masc", "measure"),
    "млн₽": ("миллион", "masc", "measure", "рублей"),
    "млрд₽": ("миллиард", "masc", "measure", "рублей"),
    "трлн₽": ("триллион", "masc", "measure", "рублей"),
    "тыс₽": ("тысяча", "femn", "measure", "рублей"),
    "млн$": ("миллион", "masc", "measure", "долларов"),
    "млрд$": ("миллиард", "masc", "measure", "долларов"),
    "тыс$": ("тысяча", "femn", "measure", "долларов"),
    "млн€": ("миллион", "masc", "measure", "евро"),
    "тыс€": ("тысяча", "femn", "measure", "евро"),
    "mah": ("миллиампер-час", "masc", "measure"),
    "мАч": ("миллиампер-час", "masc", "measure"),
    "ah": ("ампер-час", "masc", "measure"),
    "wh": ("ватт-час", "masc", "measure"),
    "kwh": ("киловатт-час", "masc", "measure"),
    "mwh": ("мегаватт-час", "masc", "measure"),
    "gwh": ("гигаватт-час", "masc", "measure"),
    "va": ("вольт-ампер", "masc", "measure"),
    "kva": ("киловольт-ампер", "masc", "measure"),
    "mol": ("моль", "femn", "measure"),
    "моль": ("моль", "femn", "measure"),
    "ммоль": ("миллимоль", "masc", "measure"),
    "mmol": ("миллимоль", "masc", "measure"),
    "мкмоль": ("микромоль", "masc", "measure"),
    "µmol": ("микромоль", "masc", "measure"),
    "μmol": ("микромоль", "masc", "measure"),
    "ммоль/л": ("миллимоль", "masc", "measure", "на литр"),
    "ммольл": ("миллимоль", "masc", "measure", "на литр"),
    "мг/л": ("миллиграмм", "masc", "measure", "на литр"),
    "мг/мл": ("миллиграмм", "masc", "measure", "на миллилитр"),
    "мкг/мл": ("микрограмм", "masc", "measure", "на миллилитр"),
    "мг/дл": ("миллиграмм", "masc", "measure", "на децилитр"),
    "ед": ("единица", "femn", "measure"),
    "ед.": ("единица", "femn", "measure"),
    "ме": ("международная единица", "femn", "measure"),
    "iu": ("международная единица", "femn", "measure"),
    "ui": ("международная единица", "femn", "measure"),
    "рад.": ("радиан", "masc", "measure"),
    "кадров/с": ("кадр", "masc", "measure", "в секунду"),
    "атм": ("атмосфер", "femn", "measure"),
}

ENTITY_KEYWORDS = {"percent": set(), "money": set(), "measure": set()}
for key, val in UNITS_DATA.items():
    lemma, _, category, *suffix = val
    ENTITY_KEYWORDS[category].add(key)
    ENTITY_KEYWORDS[category].add(lemma)

ENTITY_DEFAULT_CASE = {"percent": "nomn", "money": "nomn", "measure": "nomn"}
ALL_UNITS = set(UNITS_DATA.keys()) | {value[0] for value in UNITS_DATA.values()}

UNIT_TOKEN_FRAGMENT = (
    r"[а-яА-ЯёЁa-zA-ZµμΩ%°$€₽]+"
    r"(?:[./^][а-яА-ЯёЁa-zA-ZµμΩ%°$€₽]+|\d+|[²³])*"
)

NUMERIC_UNIT_RANGE_PATTERN = re.compile(
    rf"(?<!\d)(?P<left>\d+)\s*[—–-]\s*(?P<right>\d+)\s*(?P<unit>{UNIT_TOKEN_FRAGMENT}(?:\s+[^\W\d_]+){{0,2}})(?!\w)"
)
DECIMAL_PATTERN = re.compile(
    rf"(?<!\d)(?P<num>(?:-|\\ue001)?\d+[.,]\d+)(?:\s*(?P<unit>{UNIT_TOKEN_FRAGMENT})(?P<unit_dot>\.)?)?(?:\s+(?P<unit2>{UNIT_TOKEN_FRAGMENT})(?P<unit2_dot>\.)?)?(?!\d)"
)
FRACTION_PATTERN = re.compile(r"(?<![\d,])(\d+)/(\d+)(?![\d,])")
HYPHENATED_WORD_PATTERN = re.compile(r"(?<![\d])(\d+)-([а-яА-ЯёЁ]{1,})")
ORDINAL_PATTERN = re.compile(r"(\d+)-([а-яА-ЯёЁ]{1,3})\b", re.IGNORECASE | re.UNICODE)
STANDALONE_CURRENCY_PATTERN = re.compile(r"(?<!\S)([$€£¥₽₴₸₺₹¢])(?!\S)")
DIGIT_PATTERN = re.compile(r"\d+")

TIME_WORDS = {
    "год": "masc",
    "года": "masc",
    "лет": "masc",
    "году": "masc",
    "годом": "masc",
    "годе": "masc",
    "годов": "masc",
    "годам": "masc",
    "годами": "masc",
    "годах": "masc",
    "месяц": "masc",
    "месяца": "masc",
    "месяцев": "masc",
    "месяцу": "masc",
    "месяцем": "masc",
    "месяце": "masc",
    "месяцам": "masc",
    "месяцами": "masc",
    "месяцах": "masc",
    "день": "masc",
    "дня": "masc",
    "дней": "masc",
    "дню": "masc",
    "днем": "masc",
    "днём": "masc",
    "дне": "masc",
    "дням": "masc",
    "днями": "masc",
    "днях": "masc",
    "неделя": "femn",
    "недели": "femn",
    "недель": "femn",
    "неделе": "femn",
    "неделю": "femn",
    "неделей": "femn",
    "неделях": "femn",
    "неделям": "femn",
    "неделями": "femn",
    "сутки": "plur",
    "сут": "plur",
    "суток": "plur",
    "суткам": "plur",
    "сутками": "plur",
    "сутках": "plur",
    "час": "masc",
    "часа": "masc",
    "часов": "masc",
    "часу": "masc",
    "часом": "masc",
    "часе": "masc",
    "часам": "masc",
    "часами": "masc",
    "часах": "masc",
    "минута": "femn",
    "минуты": "femn",
    "минут": "femn",
    "минуте": "femn",
    "минуту": "femn",
    "минутой": "femn",
    "минутам": "femn",
    "минутами": "femn",
    "минутах": "femn",
    "секунда": "femn",
    "сек": "femn",
    "секунд": "femn",
    "секунды": "femn",
    "секунде": "femn",
    "секунду": "femn",
    "секундой": "femn",
    "секундам": "femn",
    "секундами": "femn",
    "секундах": "femn",
    "ч": "masc",
    "мин": "femn",
    "с": "femn",
    "дн": "masc",
    "нед": "femn",
    "в": "masc",
    "в.": "masc",
    "век": "masc",
    "века": "masc",
    "веку": "masc",
    "веке": "masc",
    "веком": "masc",
    "веков": "masc",
    "столетие": "neut",
    "столетия": "neut",
    "столетию": "neut",
    "столетии": "neut",
    "столетием": "neut",
}



from __future__ import annotations

import re
from functools import lru_cache
from pathlib import Path

from .dictionary import DictionaryNormalizer

DEFAULT_DICTIONARIES_PATH = Path(__file__).resolve().parent / "dictionaries"
DEFAULT_LATINIZATION_DICTIONARIES_PATH = DEFAULT_DICTIONARIES_PATH / "latinization"

# Order matters.
IPA_MAP = [
    ("wɝr", "вор"),
    ("wɜːr", "вор"),
    ("wɚr", "вор"),
    ("wɝ", "вор"),
    ("wɜː", "вор"),
    ("wɜ", "вор"),
    ("wɚ", "вор"),
    ("ʧ", "ч"),
    ("ʤ", "дж"),
    ("tʃ", "ч"),
    ("dʒ", "дж"),
    ("aɪ", "ай"),
    ("eɪ", "ей"),
    ("ɔɪ", "ой"),
    ("aʊ", "ау"),
    ("oʊ", "оу"),
    ("əʊ", "оу"),
    ("ju", "ю"),
    ("jʊ", "ю"),
    ("ɪə", "иэ"),
    ("eə", "эа"),
    ("ʊə", "уэ"),
    ("ɝr", "эр"),
    ("ɝ", "эр"),
    ("ɜːr", "эр"),
    ("ɜː", "эр"),
    ("ɚ", "эр"),
    ("ɾ", "т"),
    ("ŋ", "н"),
    ("ʃ", "ш"),
    ("ʒ", "ж"),
    ("θ", "с"),
    ("ð", "з"),
    ("ə", "э"),
    ("ɜ", "э"),
    ("e", "е"),
    ("æ", "э"),
    ("ɪ", "и"),
    ("ɛ", "э"),
    ("i", "и"),
    ("ɒ", "о"),
    ("ɔ", "о"),
    ("ʊ", "у"),
    ("u", "у"),
    ("ʌ", "а"),
    ("a", "а"),
    ("ɑ", "а"),
    ("p", "п"),
    ("b", "б"),
    ("t", "т"),
    ("d", "д"),
    ("k", "к"),
    ("g", "г"),
    ("f", "ф"),
    ("v", "в"),
    ("s", "с"),
    ("z", "з"),
    ("h", "х"),
    ("m", "м"),
    ("n", "н"),
    ("l", "л"),
    ("r", "р"),
    ("j", "й"),
    ("w", "в"),
    ("ʔ", "т"),
    ("ˌ", ""),
]


@lru_cache(maxsize=16)
def _get_latin_dictionary_normalizer(
    dictionaries_path: str, filename: str
) -> DictionaryNormalizer:
    return DictionaryNormalizer(
        dictionaries_path=dictionaries_path,
        include_only_files=[filename],
    )


@lru_cache(maxsize=50000)
def _ipa_convert_cached(word: str) -> str:
    import eng_to_ipa as ipa

    return ipa.convert(word)


@lru_cache(maxsize=50000)
def _resolve_unknown_latin_fallback(
    word: str, dictionaries_path: str, filename: str
) -> str:
    current = word
    resolved_path = Path(dictionaries_path)
    for _ in range(6):
        updated = _apply_dictionary_latinization(current, resolved_path, filename)
        if updated == current:
            break
        current = updated
    return current


def handle_long_vowels(ipa: str) -> str:
    long_vowels = {
        "iː": "ii",
        "uː": "uu",
        "ɑː": "aa",
        "ɔː": "oo",
    }
    for old, new in long_vowels.items():
        ipa = ipa.replace(old, new)
    return ipa


def move_stress_marker_ru(text: str) -> str:
    vowels = set("аеёиоуыэюя")
    result: list[str] = []
    pending = False

    for char in text:
        if char == "ˈ":
            pending = True
            continue
        if pending and char in vowels:
            result.append("ˈ")
            pending = False
        result.append(char)

    return "".join(result)


def _ipa_to_russian(ipa_text: str, include_stress_markers: bool = False) -> str:
    ipa_text = ipa_text.replace("ˌ", "")
    ipa_text = handle_long_vowels(ipa_text)
    for ipa_char, ru_char in IPA_MAP:
        ipa_text = ipa_text.replace(ipa_char, ru_char)
    ipa_text = ipa_text.replace("ː", "")
    ipa_text = move_stress_marker_ru(ipa_text)
    return ipa_text.replace("ˈ", "+" if include_stress_markers else "")


def _apply_dictionary_latinization(
    text: str, dictionaries_path: Path, filename: str
) -> str:
    normalizer = _get_latin_dictionary_normalizer(str(dictionaries_path), filename)
    return normalizer.apply(text, strip_unmatched_latin=False)


def _resolve_latinization_dictionary_source(
    dictionaries_path: Path | None, filename: str
) -> tuple[Path, str]:
    if dictionaries_path is None:
        default_file = DEFAULT_LATINIZATION_DICTIONARIES_PATH / filename
        if default_file.exists():
            return DEFAULT_LATINIZATION_DICTIONARIES_PATH, filename
        return DEFAULT_LATINIZATION_DICTIONARIES_PATH, "latinization_rules.dic"

    direct_path = dictionaries_path / filename
    if direct_path.exists():
        return dictionaries_path, filename

    nested_path = dictionaries_path / "latinization"
    nested_file = nested_path / filename
    if nested_file.exists():
        return nested_path, filename

    default_file = DEFAULT_LATINIZATION_DICTIONARIES_PATH / filename
    if default_file.exists():
        return DEFAULT_LATINIZATION_DICTIONARIES_PATH, filename

    bundled_default_filename = "latinization_rules.dic"
    bundled_default_file = (
        DEFAULT_LATINIZATION_DICTIONARIES_PATH / bundled_default_filename
    )
    if bundled_default_file.exists():
        return DEFAULT_LATINIZATION_DICTIONARIES_PATH, bundled_default_filename

    return dictionaries_path, filename


def _apply_ipa_latinization(
    text: str,
    dictionaries_path: Path,
    filename: str,
    *,
    include_stress_markers: bool,
) -> str:
    try:
        import eng_to_ipa  # noqa: F401
    except ImportError:
        return text

    def replace(match: re.Match[str]) -> str:
        word = match.group(0)
        ipa_text = _ipa_convert_cached(word.lower())
        if "*" in ipa_text:
            return _resolve_unknown_latin_fallback(
                word, str(dictionaries_path), filename
            )
        return _ipa_to_russian(
            ipa_text, include_stress_markers=include_stress_markers
        )

    return re.sub(r"[A-Za-z][A-Za-z'\-]*", replace, text)


def apply_latinization(
    text: str,
    *,
    enabled: bool,
    backend: str,
    dictionaries_path: Path | None = None,
    dictionary_filename: str = "latinization_rules.dic",
    include_stress_markers: bool = False,
) -> str:
    if not enabled or not re.search(r"[A-Za-z]", text):
        return text

    dict_path, resolved_filename = _resolve_latinization_dictionary_source(
        dictionaries_path, dictionary_filename
    )
    backend_name = backend.lower()
    if backend_name == "dictionary":
        return _apply_dictionary_latinization(text, dict_path, resolved_filename)
    if backend_name == "ipa":
        return _apply_ipa_latinization(
            text,
            dict_path,
            resolved_filename,
            include_stress_markers=include_stress_markers,
        )
    return text



from __future__ import annotations

import re

from .constants import CLEANUP_REPLACEMENTS, UNICODE_FRACTIONS

ASCII_DASH_SPACE_PATTERN = re.compile(r" - ")
LETTER_HYPHEN_PATTERN = re.compile(r"(?<=[A-Za-zА-Яа-яЁё])-(?=[A-Za-zА-Яа-яЁё])")
SLASH_FIX_PATTERN = re.compile(r"(?<=[a-zA-Zа-яА-ЯёЁ+])/(?=[a-zA-Zа-яА-ЯёЁ+])")
UNIT_SLASH_PATTERN = re.compile(
    r"\b(?P<left>об|кадров|[kmg]?b|[кмг]?бит|[кмг]?б|байт|[км]?моль|мг|мкг|км|м)/(?P<right>s|с|ч|мин|л|мл|дл)\b",
    re.IGNORECASE,
)
NUMBER_CLEANUP_PATTERN = re.compile(
    r"(?<!\d)(?:\d{1,3}(?:[ \u00A0\u2009\u202F]\d{3})+|\d+)(?:[,\.]\d+)?(?!\d)"
)
BRACKETED_NUMERIC_PATTERN = re.compile(r"([\[<({«])\s*([^()\[\]{}<>«»]+?)\s*([\]>)}»])")
SPACE_BEFORE_PUNCT_PATTERN = re.compile(r"[ \t]+([,.;:!?])")
MULTI_SPACE_PATTERN = re.compile(r"[ \t]{2,}")
SPACE_BEFORE_CLOSE_BRACKET_PATTERN = re.compile(r"[ \t]+([\)\]\}])")
SPACE_AFTER_OPEN_BRACKET_PATTERN = re.compile(r"([\(\[\{])[ \t]+")
TRAILING_SPACE_BEFORE_NEWLINE_PATTERN = re.compile(r"[ \t]+\n")
LEADING_SPACE_AFTER_NEWLINE_PATTERN = re.compile(r"\n[ \t]+")
EXCESSIVE_LINEBREAKS_PATTERN = re.compile(r"\n{2,}")
DECORATIVE_SEPARATOR_PATTERN = re.compile(
    r"(?:(?<=^)|(?<=\n))[ \t]*(?:[*=_~+#\-\xad\u2010-\u2015][ \t]*){3,}(?:(?=$)|(?=\n))"
)
LEGACY_PARAGRAPH_BREAK_DOT_PATTERN = re.compile(
    r"(?P<prev>[^\n.!?…,:;–\-\"'])(?P<break>\n{2,})(?=(?:[\"«„“]\s*)?[А-ЯЁA-Z0-9])"
)
LEGACY_LINEBREAK_DOT_PATTERN = re.compile(
    r"(?P<prev>[^\n.!?…,:;–\-\"'])(?P<break>\n)(?=(?:[\"«„“]\s*)?[А-ЯЁA-Z0-9])"
)
INLINE_LINEBREAK_SPACE_PATTERN = re.compile(
    r"(?<=[^\n])\n(?!(?:[\"«„“]\s*)?[А-ЯЁA-Z0-9])(?=[^\n])"
)
LETTER_HYPHEN_PLACEHOLDER = "\ue000"
NEGATIVE_NUMBER_PLACEHOLDER = "\ue001"
PARAGRAPH_BREAK_PLACEHOLDER = "\ue002"
UNIT_SLASH_PLACEHOLDER = "\ue003"


def protect_unit_slashes(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        return (
            f"{match.group('left')}{UNIT_SLASH_PLACEHOLDER}{match.group('right')}"
        )

    return UNIT_SLASH_PATTERN.sub(repl, text)
YEARS_AGO_ABBREVIATION_PATTERN = re.compile(
    r"(?<!\w)л\.?\s*н\.(?P<tail>\s*)",
    re.IGNORECASE,
)
UNARY_MINUS_NUMBER_PATTERN = re.compile(
    r"(?P<prefix>^|[(\[{«]|(?<!\d)\s)(?P<sign>[−-])(?P<space>\s*)(?P<number>\d+(?:[.,]\d+)?)"
)
THOUSANDS_SEPARATORS = " \u00a0\u2009\u202f"
SIGNED_COMMA_NUMBER_PATTERN = re.compile(
    r"^[−-]?(?:\d{1,3}(?:[" + THOUSANDS_SEPARATORS + r"]\d{3})+|\d+),\d+$"
)
SIGNED_INTEGER_WITH_UNIT_PATTERN = re.compile(
    r"^[−-]?(?:\d{1,3}(?:[" + THOUSANDS_SEPARATORS + r"]\d{3})+|\d+)(?:,\d+)?\s*"
    r"(?:[%‰°]|[₽$€£¥]|[A-Za-zА-Яа-яЁё]+(?:\s+[A-Za-zА-Яа-яЁё.]+)?)$"
)
INTEGER_LIST_PATTERN = re.compile(r"^\d+(?:\s*,\s*\d+)+$")
INTEGER_RANGE_PATTERN = re.compile(r"^\d+\s*[–—-]\s*\d+$")
SPACE_INSIDE_QUOTES_PATTERN = re.compile(
    r'(^|[\s([{\-–—,;:])"\s+([^\n"]+?)\s+"(?=$|[\s)\]}\-–—,.;:!?])'
)
SPACE_AFTER_OPEN_QUOTE_PATTERN = re.compile(
    r'(^|[\s([{\-–—,;:])"([ \t]+)(?=\S)'
)
SPACE_BEFORE_CLOSE_QUOTE_PATTERN = re.compile(
    r'(?<=\S)([ \t]+)"(?=$|[\s)\]}\-–—,.;:!?])'
)
OPEN_SINGLE_QUOTE_PATTERN = re.compile(
    r"(?P<prefix>^|[\s([{\-–—,;:])(?P<quote>[‘‚‛‹])(?=\S)"
)
CLOSE_SINGLE_QUOTE_PATTERN = re.compile(
    r"(?<=\S)(?P<quote>[’›])(?=$|[\s)\]}\-–—,.;:!?])"
)
ELLIPSIS_SPACE_BEFORE_PATTERN = re.compile(r"[ \t]+(?=…)")
ELLIPSIS_SPACE_AFTER_PATTERN = re.compile(r"(?<=…)(?=[^\s.,;:!?…)\]}\"])\S")
SENTENCE_SPACE_AFTER_PATTERN = re.compile(r"(?<=[.!?…])(?=[\"(«„“]?[A-ZА-ЯЁ])")
CYRILLIC_COMBINING_STRESS_PATTERN = re.compile(
    r"([АЕЁИОУЫЭЮЯаеёиоуыэюя])([\u0300\u0301])"
)
ZERO_WIDTH_FORMATTING_PATTERN = re.compile(r"[\u200B\u200C\u200D\u2060\uFEFF]")
PAGE_ABBREVIATION_PATTERN = re.compile(r"\b[сp]\.\s*(?=\d)", re.IGNORECASE)
APPROXIMATE_ABBREVIATION_PATTERN = re.compile(r"\bок\.\s*(?=\d)", re.IGNORECASE)
ERA_ABBREVIATION_PATTERN = re.compile(
    r"(?<!\w)(?P<abbr>до\s+н\.?\s*э\.?|н\.?\s*э\.?)(?P<tail>\s*)",
    re.IGNORECASE,
)


def normalize_ascii_quote_pairs(text: str) -> str:
    replacements = (
        ("``", '"'),
        ("''", '"'),
        ("‘‘", '"'),
        ("’’", '"'),
        ("´´", '"'),
        ("«", '"'),
        ("»", '"'),
        ("“", '"'),
        ("”", '"'),
        ("„", '"'),
        ("‟", '"'),
        ("〝", '"'),
        ("〞", '"'),
        ("＂", '"'),
    )
    for old, new in replacements:
        text = text.replace(old, new)
    text = OPEN_SINGLE_QUOTE_PATTERN.sub(r'\g<prefix>"', text)
    text = CLOSE_SINGLE_QUOTE_PATTERN.sub('"', text)
    text = SPACE_INSIDE_QUOTES_PATTERN.sub(r'\1"\2"', text)
    text = SPACE_AFTER_OPEN_QUOTE_PATTERN.sub(r'\1"', text)
    return SPACE_BEFORE_CLOSE_QUOTE_PATTERN.sub('"', text)


def normalize_punctuation_spacing(text: str) -> str:
    text = ELLIPSIS_SPACE_BEFORE_PATTERN.sub("", text)
    text = ELLIPSIS_SPACE_AFTER_PATTERN.sub(lambda m: f" {m.group(0)}", text)
    return SENTENCE_SPACE_AFTER_PATTERN.sub(" ", text)


def normalize_cyrillic_combining_stress_marks(text: str) -> str:
    return CYRILLIC_COMBINING_STRESS_PATTERN.sub(r"+\1", text)


def remove_zero_width_formatting(text: str) -> str:
    return ZERO_WIDTH_FORMATTING_PATTERN.sub("", text)


def normalize_era_abbreviations(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        abbr = match.group("abbr")
        tail = match.group("tail")
        expansion = "до нашей эры" if abbr.lower().startswith("до") else "нашей эры"
        rest = text[match.end() :]
        stripped_rest = rest.lstrip()
        keep_terminal_dot = False
        if abbr.rstrip().endswith("."):
            keep_terminal_dot = (
                not stripped_rest
                or "\n" in tail
                or (stripped_rest[:1] and stripped_rest[:1].isupper())
            )
            if (
                stripped_rest[:1] in '"»”)]}'
                and len(stripped_rest) > 1
                and stripped_rest[1:2].isupper()
            ):
                keep_terminal_dot = True
        return f"{expansion}{'.' if keep_terminal_dot else ''}{tail}"

    return ERA_ABBREVIATION_PATTERN.sub(repl, text)


def normalize_numeric_abbreviations(text: str) -> str:
    text = PAGE_ABBREVIATION_PATTERN.sub("страница ", text)
    text = APPROXIMATE_ABBREVIATION_PATTERN.sub("около ", text)
    return normalize_era_abbreviations(text)


def normalize_explicit_dashes(text: str) -> str:
    return text.translate(str.maketrans({"–": "—", "―": "—"}))


def normalize_spaced_ascii_hyphens(text: str) -> str:
    return ASCII_DASH_SPACE_PATTERN.sub(" — ", text)


def protect_negative_numbers(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        prefix = match.group("prefix")
        number = match.group("number")
        return f"{prefix}{NEGATIVE_NUMBER_PLACEHOLDER}{number}"

    return UNARY_MINUS_NUMBER_PATTERN.sub(repl, text)


def protect_letter_hyphens(text: str) -> str:
    return LETTER_HYPHEN_PATTERN.sub(LETTER_HYPHEN_PLACEHOLDER, text)


def restore_letter_hyphens(text: str) -> str:
    return text.replace(LETTER_HYPHEN_PLACEHOLDER, "-")


def expand_years_ago_abbreviation(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        tail = match.group("tail")
        next_char_index = match.end()
        next_char = text[next_char_index] if next_char_index < len(text) else ""
        if (
            not next_char
            or next_char == "\n"
            or next_char.isalnum()
            or next_char.isalpha()
        ):
            return f"лет назад.{tail}"
        return f"лет назад{tail}"

    return YEARS_AGO_ABBREVIATION_PATTERN.sub(repl, text)


def normalize_unicode_fractions(text: str) -> str:
    for char, replacement in UNICODE_FRACTIONS.items():
        text = text.replace(char, replacement)
    return text


def apply_cleanup_replacements(text: str) -> str:
    for old, new in CLEANUP_REPLACEMENTS:
        text = text.replace(old, new)
    return text


def restore_paragraph_breaks(text: str) -> str:
    return text


def _insert_boundary_dot(match: re.Match[str]) -> str:
    return f"{match.group('prev')}.{match.group('break')}"


def normalize_linebreaks(text: str, keep_paragraph_placeholders: bool = False) -> str:
    del keep_paragraph_placeholders
    text = text.replace("\r\n", "\n").replace("\r", "\n")
    text = TRAILING_SPACE_BEFORE_NEWLINE_PATTERN.sub("\n", text)
    text = LEADING_SPACE_AFTER_NEWLINE_PATTERN.sub("\n", text)
    text = EXCESSIVE_LINEBREAKS_PATTERN.sub("\n\n", text)
    text = LEGACY_PARAGRAPH_BREAK_DOT_PATTERN.sub(_insert_boundary_dot, text)
    text = LEGACY_LINEBREAK_DOT_PATTERN.sub(_insert_boundary_dot, text)
    text = INLINE_LINEBREAK_SPACE_PATTERN.sub(" ", text)
    return text


def remove_decorative_separators(text: str) -> str:
    return DECORATIVE_SEPARATOR_PATTERN.sub("", text)


def classify_bracketed_numeric_content(content: str) -> str:
    stripped = content.strip()
    if not stripped:
        return "other"

    if SIGNED_COMMA_NUMBER_PATTERN.fullmatch(stripped):
        return "numeric"

    if SIGNED_INTEGER_WITH_UNIT_PATTERN.fullmatch(stripped):
        return "numeric"

    if stripped.isdigit():
        return "reference"

    if INTEGER_RANGE_PATTERN.fullmatch(stripped):
        return "reference"

    if INTEGER_LIST_PATTERN.fullmatch(stripped):
        if stripped.count(",") == 1 and ", " not in stripped and " ," not in stripped:
            return "numeric"
        return "reference"

    if "." in stripped:
        parts = stripped.split(".")
        if all(part.isdigit() for part in parts):
            if len(parts) >= 3:
                return "reference"
            if len(parts) == 2:
                left, right = parts
                if len(left) <= 2 and len(right) <= 3:
                    return "reference"
                return "numeric"

    return "other"


def remove_numeric_footnotes(
    text: str,
    keep_paragraph_placeholders: bool = False,
    ignore_interval: tuple[int, int] | None = None,
) -> str:
    pairs = {"[": "]", "<": ">", "(": ")", "{": "}", "«": "»"}

    def repl(match: re.Match[str]) -> str:
        opener, content, closer = match.group(1), match.group(2), match.group(3)
        if pairs.get(opener) != closer:
            return match.group(0)
        stripped = content.strip()
        if classify_bracketed_numeric_content(content) == "reference":
            if ignore_interval is not None and stripped.isdigit():
                value = int(stripped)
                min_val, max_val = ignore_interval
                if min_val <= value <= max_val:
                    return match.group(0)
            return ""
        return match.group(0)

    text = BRACKETED_NUMERIC_PATTERN.sub(repl, text)
    text = SPACE_BEFORE_PUNCT_PATTERN.sub(r"\1", text)
    text = MULTI_SPACE_PATTERN.sub(" ", text)
    text = SPACE_BEFORE_CLOSE_BRACKET_PATTERN.sub(r"\1", text)
    text = SPACE_AFTER_OPEN_BRACKET_PATTERN.sub(r"\1", text)
    return normalize_linebreaks(
        text, keep_paragraph_placeholders=keep_paragraph_placeholders
    ).strip()


def clean_numbers(text: str) -> str:
    def repl_spaces(match: re.Match[str]) -> str:
        return (
            match.group(0)
            .replace(" ", "")
            .replace("\u00a0", "")
            .replace("\u2009", "")
            .replace("\u202f", "")
        )

    text = NUMBER_CLEANUP_PATTERN.sub(repl_spaces, text)
    text = re.sub(
        r"(?<!\d)(\d{1,3}(?:\.\d{3}){2,})(?!\d)",
        lambda m: m.group(1).replace(".", ""),
        text,
    )
    text = re.sub(
        r"(?<!\d)(\d{1,3}\.\d{3})(?=,\d+)", lambda m: m.group(1).replace(".", ""), text
    )
    unit_lookahead = (
        r"(?=\s*(?:[₽$€£¥%‰]|\b(?:руб|долл|евро|тыс|млн|млрд|г|кг|т|м|км|шт)\b))"
    )
    return re.sub(
        r"(?<!\d)(\d{1,3}\.\d{3})" + unit_lookahead,
        lambda m: m.group(1).replace(".", ""),
        text,
    )



from __future__ import annotations

import re

from .preprocess_utils import NEGATIVE_NUMBER_PLACEHOLDER, PARAGRAPH_BREAK_PLACEHOLDER

TOKEN_PATTERN = re.compile(
    rf"\n+|{re.escape(PARAGRAPH_BREAK_PLACEHOLDER)}+|{re.escape(NEGATIVE_NUMBER_PLACEHOLDER)}\d+(?:[.,]\d+)?|\d+-[A-Za-z]+|[^\W\d_]+-\d+|\d+|\+?[^\W\d_]+(?:[-+][^\W\d_]+)+|\+?[^\W\d_]+|[$€₽£¥₴₸₺₹¢₪₩₫₱₦]|[^\w\s]",
    flags=re.UNICODE,
)
PUNCT_STRIP = '.,:;!"«»()[]{}'


def simple_tokenize(text: str) -> list[str]:
    return TOKEN_PATTERN.findall(text)


def normalize_context_token(token: str) -> str:
    return token.lower().strip(PUNCT_STRIP)
